# Copyright (C) 2018-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Utility module for browsing the archive in a snapshot context.

from collections import defaultdict, deque
import dataclasses
from typing import Any, Dict, List, Optional, Tuple

from looseversion import LooseVersion2

from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.utils.html import escape

from swh.model.hashutil import hash_to_bytes
from swh.model.model import Snapshot
from swh.model.swhids import CoreSWHID, ObjectType
from swh.storage.algos.revisions_walker import State
from swh.web.browse.utils import (
    format_log_entries,
    gen_release_link,
    gen_revision_link,
    gen_revision_log_link,
    gen_revision_url,
    gen_snapshot_link,
    get_directory_entries,
    get_readme_to_display,
)
from swh.web.config import get_config
from swh.web.utils import (
    archive,
    cache_get,
    cache_set,
    django_cache,
    format_utc_iso_date,
    gen_path_info,
    reverse,
    swh_object_icons,
)
from swh.web.utils.exc import BadInputExc, NotFoundExc, http_status_code_message
from swh.web.utils.identifiers import get_swhids_info
from swh.web.utils.origin_visits import get_origin_visit
from swh.web.utils.typing import (
    DirectoryMetadata,
    OriginInfo,
    SnapshotBranchInfo,
    SnapshotContext,
    SnapshotReleaseInfo,
    SWHObjectInfo,
)

_empty_snapshot_id = Snapshot(branches={}).id.hex()


def _get_branch(
    branches: List[SnapshotBranchInfo], branch_name: str, snapshot_id: str
) -> Optional[SnapshotBranchInfo]:
    """
    Utility function to get a specific branch from a snapshot.
    Returns None if the branch cannot be found.
    """
    filtered_branches = [b for b in branches if b["name"] == branch_name]
    if filtered_branches:
        return filtered_branches[0]
    else:
        # case where a large branches list has been truncated
        snp = archive.lookup_snapshot(
            snapshot_id,
            branches_from=branch_name,
            branches_count=1,
            target_types=["revision", "alias", "content", "directory"],
            # pull request branches must be browsable even if they are hidden
            # by default in branches list
            branch_name_exclude_prefix=None,
        )
        snp_branch, _, _ = process_snapshot_branches(snp)
        if snp_branch and snp_branch[0]["name"] == branch_name:
            branches.append(snp_branch[0])
            return snp_branch[0]
    return None


def _get_release(
    releases: List[SnapshotReleaseInfo], release_name: Optional[str], snapshot_id: str
) -> Optional[SnapshotReleaseInfo]:
    """
    Utility function to get a specific release from a snapshot.
    Returns None if the release cannot be found.
    """
    filtered_releases = [r for r in releases if r["name"] == release_name]
    if filtered_releases:
        return filtered_releases[0]
    elif release_name:
        # case where a large branches list has been truncated

        # git/hg/package origins have specific branches for releases
        for branch_prefix in ("refs/tags", "tags", "releases"):
            try:
                branch_name = f"{branch_prefix}/{release_name}"
                snp = archive.lookup_snapshot(
                    snapshot_id,
                    branches_from=branch_name,
                    branches_count=1,
                    target_types=["release", "alias"],
                )
                if branch_name in snp["branches"]:
                    break
            except NotFoundExc:
                pass
        else:
            snp = archive.lookup_snapshot(
                snapshot_id,
                branches_from=release_name,
                branches_count=1,
                target_types=["release", "alias"],
            )
        _, snp_release, _ = process_snapshot_branches(snp)
        if snp_release and snp_release[0]["name"] == release_name:
            releases.append(snp_release[0])
            return snp_release[0]
    return None


def _branch_not_found(
    branch_type: str,
    branch: str,
    snapshot_id: str,
    snapshot_sizes: Dict[str, int],
    origin_info: Optional[OriginInfo],
    timestamp: Optional[str],
    visit_id: Optional[int],
) -> None:
    """
    Utility function to raise an exception when a specified branch/release
    cannot be found.
    """
    if branch_type == "branch":
        branch_type = "Branch"
        branch_type_plural = "branches"
        target_type = "branch"
    else:
        branch_type = "Release"
        branch_type_plural = "releases"
        target_type = "release"

    if snapshot_id and snapshot_sizes[target_type] == 0:
        msg = "Snapshot with id %s has an empty list" " of %s!" % (
            snapshot_id,
            branch_type_plural,
        )
    elif snapshot_id:
        msg = "%s %s for snapshot with id %s" " not found!" % (
            branch_type,
            branch,
            snapshot_id,
        )
    elif visit_id and snapshot_sizes[target_type] == 0 and origin_info:
        msg = (
            "Origin with url %s"
            " for visit with id %s has an empty list"
            " of %s!" % (origin_info["url"], visit_id, branch_type_plural)
        )
    elif visit_id and origin_info:
        msg = (
            "%s %s associated to visit with"
            " id %s for origin with url %s"
            " not found!" % (branch_type, branch, visit_id, origin_info["url"])
        )
    elif snapshot_sizes[target_type] == 0 and origin_info and timestamp:
        msg = (
            "Origin with url %s"
            " for visit with timestamp %s has an empty list"
            " of %s!" % (origin_info["url"], timestamp, branch_type_plural)
        )
    elif origin_info and timestamp:
        msg = (
            "%s %s associated to visit with"
            " timestamp %s for origin with "
            "url %s not found!" % (branch_type, branch, timestamp, origin_info["url"])
        )
    raise NotFoundExc(escape(msg))


def process_snapshot_branches(
    snapshot: Dict[str, Any], discard_messages: bool = False
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo], Dict[str, Any]]:
    """
    Process a dictionary describing snapshot branches: extract those
    targeting revisions and releases, put them in two different lists,
    then sort those lists in lexicographical order of the branches' names.

    Args:
        snapshot: a dict describing a snapshot as returned for instance by
            :func:`swh.web.utils.archive.lookup_snapshot`
        discard_messages: if :const:`True`, do not retrieve messages associated
            to branches and releases

    Returns:
        A tuple whose first member is the sorted list of branches
        targeting revisions, second member the sorted list of branches
        targeting releases and third member a dict mapping resolved branch
        aliases to their real target.
    """
    snapshot_branches = snapshot["branches"]
    branches: Dict[str, SnapshotBranchInfo] = {}
    branch_aliases: Dict[str, str] = {}
    releases: Dict[str, SnapshotReleaseInfo] = {}
    revision_to_branch = defaultdict(set)
    revision_to_release = defaultdict(set)
    release_to_branch = defaultdict(set)
    for branch_name, target in snapshot_branches.items():
        if not target:
            # FIXME: display branches with an unknown target anyway
            continue
        target_id = target["target"]
        target_type = target["target_type"]
        if target_type in ("content", "directory", "revision"):
            branches[branch_name] = SnapshotBranchInfo(
                name=branch_name,
                alias=False,
                target_type=target_type,
                target=target_id,
                date=None,
                directory=None,
                message=None,
                url=None,
            )
            if target_type == "revision":
                revision_to_branch[target_id].add(branch_name)
        elif target_type == "release":
            release_to_branch[target_id].add(branch_name)
        elif target_type == "alias":
            branch_aliases[branch_name] = target_id
        # FIXME: handle pointers to other object types

    def _add_release_info(branch, release, alias=False):
        releases[branch] = SnapshotReleaseInfo(
            name=release["name"],
            alias=alias,
            branch_name=branch,
            date=format_utc_iso_date(release["date"]),
            directory=None,
            id=release["id"],
            message=release["message"] if not discard_messages else None,
            target_type=release["target_type"],
            target=release["target"],
            url=None,
        )

    def _add_branch_info(branch, revision, alias=False):
        branches[branch] = SnapshotBranchInfo(
            name=branch,
            alias=alias,
            target_type="revision",
            target=revision["id"],
            directory=revision["directory"],
            date=format_utc_iso_date(revision["date"]),
            message=revision["message"] if not discard_messages else None,
            url=None,
        )

    releases_info = archive.lookup_release_multiple(release_to_branch.keys())
    for release in releases_info:
        if release is None:
            continue
        branches_to_update = release_to_branch[release["id"]]
        for branch in branches_to_update:
            _add_release_info(branch, release)
        if release["target_type"] == "revision":
            revision_to_release[release["target"]].update(branches_to_update)

    revisions = archive.lookup_revision_multiple(
        set(revision_to_branch.keys()) | set(revision_to_release.keys())
    )

    for revision in revisions:
        if not revision:
            continue
        for branch in revision_to_branch[revision["id"]]:
            _add_branch_info(branch, revision)
        for release_id in revision_to_release[revision["id"]]:
            releases[release_id]["directory"] = revision["directory"]

    resolved_aliases = {}

    for branch_alias, _ in branch_aliases.items():
        resolved_alias = archive.lookup_snapshot_alias(snapshot["id"], branch_alias)
        resolved_aliases[branch_alias] = resolved_alias
        if resolved_alias is None:
            continue

        target_type = resolved_alias["target_type"]
        target = resolved_alias["target"]

        if target_type == "revision":
            revision = archive.lookup_revision(target)
            _add_branch_info(branch_alias, revision, alias=True)
        elif target_type == "release":
            release = archive.lookup_release(target)
            _add_release_info(branch_alias, release, alias=True)
        elif target_type in ("content", "directory"):
            branches[branch_name] = SnapshotBranchInfo(
                name=branch_alias,
                alias=True,
                target_type=target_type,
                target=target,
                date=None,
                directory=None,
                message=None,
                url=None,
            )

        if branch_alias in branches:
            branches[branch_alias]["name"] = branch_alias

    ret_branches = list(sorted(branches.values(), key=lambda b: b["name"]))
    ret_releases = list(
        reversed(sorted(releases.values(), key=lambda b: LooseVersion2(b["name"])))
    )

    return ret_branches, ret_releases, resolved_aliases


@django_cache()
def get_snapshot_content(
    snapshot_id: str,
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo], Dict[str, Any]]:
    """Returns the lists of branches and releases
    associated to a swh snapshot.
    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        snapshot_id: hexadecimal representation of the snapshot identifier

    Returns:
        A tuple with three members. The first one is a list of dict describing
        the snapshot branches. The second one is a list of dict describing the
        snapshot releases. The third one is a dict mapping resolved branch
        aliases to their real target.

    Raises:
        NotFoundExc if the snapshot does not exist
    """

    branches: List[SnapshotBranchInfo] = []
    releases: List[SnapshotReleaseInfo] = []
    aliases: Dict[str, Any] = {}

    snapshot_content_max_size = get_config()["snapshot_content_max_size"]

    if snapshot_id:
        snapshot = archive.lookup_snapshot(
            snapshot_id, branches_count=snapshot_content_max_size
        )
        branches, releases, aliases = process_snapshot_branches(
            snapshot, discard_messages=True
        )

    return branches, releases, aliases


def get_origin_visit_snapshot(
    origin_info: OriginInfo,
    visit_ts: Optional[str] = None,
    visit_id: Optional[int] = None,
    snapshot_id: Optional[str] = None,
) -> Tuple[List[SnapshotBranchInfo], List[SnapshotReleaseInfo], Dict[str, Any]]:
    """Returns the lists of branches and releases associated to an origin for
    a given visit.

    The visit is expressed by either:

        * a snapshot identifier
        * a timestamp, if no visit with that exact timestamp is found,
          the closest one from the provided timestamp will be used.

    If no visit parameter is provided, it returns the list of branches
    found for the latest visit.

    That list is put in  cache in order to speedup the navigation
    in the swh-web/browse ui.

    .. warning:: At most 1000 branches contained in the snapshot
        will be returned for performance reasons.

    Args:
        origin_info: a dict filled with origin information
        visit_ts: an ISO 8601 datetime string to parse
        visit_id: visit id for disambiguation in case several visits have
            the same timestamp
        snapshot_id: if provided, visit associated to the snapshot will be processed

    Returns:
        A tuple with three members. The first one is a list of dict describing
        the origin branches for the given visit.
        The second one is a list of dict describing the origin releases
        for the given visit. The third one is a dict mapping resolved branch
        aliases to their real target.

    Raises:
        NotFoundExc if the origin or its visit are not found
    """

    visit_info = get_origin_visit(origin_info["url"], visit_ts, visit_id, snapshot_id)

    return get_snapshot_content(visit_info["snapshot"])


def get_snapshot_context(
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
    visit_id: Optional[int] = None,
    branch_name: Optional[str] = None,
    release_name: Optional[str] = None,
    revision_id: Optional[str] = None,
    path: Optional[str] = None,
    browse_context: str = "directory",
    visit_type: Optional[str] = None,
) -> SnapshotContext:
    """
    Utility function to compute relevant information when navigating
    the archive in a snapshot context. The snapshot is either
    referenced by its id or it will be retrieved from an origin visit.

    Args:
        snapshot_id: hexadecimal representation of a snapshot identifier
        origin_url: an origin_url
        timestamp: a datetime string for retrieving the closest
            visit of the origin
        visit_id: optional visit id for disambiguation in case
            of several visits with the same timestamp
        branch_name: optional branch name set when browsing the snapshot in
            that scope (will default to "HEAD" if not provided)
        release_name: optional release name set when browsing the snapshot in
            that scope
        revision_id: optional revision identifier set when browsing the snapshot in
            that scope
        path: optional path of the object currently browsed in the snapshot
        browse_context: indicates which type of object is currently browsed

    Returns:
        A dict filled with snapshot context information.

    Raises:
        swh.web.utils.exc.NotFoundExc: if no snapshot is found for the visit
            of an origin.
    """
    assert origin_url is not None or snapshot_id is not None
    origin_info = None
    visit_info = None
    url_args = {}
    query_params: Dict[str, Any] = {"visit_type": visit_type}
    origin_visits_url = None

    if origin_url:
        if visit_id is not None:
            query_params["visit_id"] = visit_id
        elif snapshot_id is not None:
            query_params["snapshot"] = snapshot_id

        origin_info = archive.lookup_origin(origin_url)

        query_params["origin_url"] = origin_info["url"]
        if (
            len(origin_info["visit_types"]) > 1
            and visit_type is None
            and snapshot_id is None
            and timestamp is None
            and visit_id is None
        ):
            # avoid selecting git-checkout, svn-export or hg-checkout visit types
            # when no visit type is explicitly provided
            for visit_type_to_discard, visit_type_to_fetch in [
                ("git-checkout", "git"),
                ("hg-checkout", "hg"),
                ("svn-export", "svn"),
            ]:
                if visit_type_to_discard in origin_info["visit_types"]:
                    visit_type = visit_type_to_fetch
                    break

        visit_info = get_origin_visit(
            origin_info["url"], timestamp, visit_id, snapshot_id, visit_type=visit_type
        )

        if len(origin_info["visit_types"]) > 1:
            query_params["visit_type"] = visit_info["type"]

        formatted_date = format_utc_iso_date(visit_info["date"])
        visit_info["formatted_date"] = formatted_date
        snapshot_id = visit_info.get("snapshot")

        if not snapshot_id:
            raise NotFoundExc(
                "No snapshot associated to the visit of origin "
                "%s on %s" % (escape(origin_url), formatted_date)
            )

        # provided timestamp is not necessarily equals to the one
        # of the retrieved visit, so get the exact one in order
        # to use it in the urls generated below
        if timestamp:
            timestamp = visit_info["date"]

        branches, releases, aliases = get_origin_visit_snapshot(
            origin_info, timestamp, visit_id, snapshot_id
        )

        origin_visits_url = reverse(
            "browse-origin-visits",
            query_params={
                "origin_url": origin_info["url"],
                "visit_type": query_params["visit_type"],
            },
        )

        if timestamp is not None:
            query_params["timestamp"] = format_utc_iso_date(
                timestamp, "%Y-%m-%dT%H:%M:%SZ"
            )

        visit_url = reverse("browse-origin-directory", query_params=query_params)
        visit_info["url"] = browse_url = visit_url

        branches_url = reverse("browse-origin-branches", query_params=query_params)

        releases_url = reverse("browse-origin-releases", query_params=query_params)
    else:
        assert snapshot_id is not None
        branches, releases, aliases = get_snapshot_content(snapshot_id)
        url_args = {"snapshot_id": snapshot_id}
        browse_url = reverse("browse-snapshot-directory", url_args=url_args)
        branches_url = reverse("browse-snapshot-branches", url_args=url_args)

        releases_url = reverse("browse-snapshot-releases", url_args=url_args)

    @django_cache()
    def _get_snapshot_sizes(snapshot_id):
        return archive.lookup_snapshot_sizes(snapshot_id)

    snapshot_sizes = _get_snapshot_sizes(snapshot_id)

    snapshot_total_size = sum(
        v for k, v in snapshot_sizes.items() if k not in ("alias", "branch")
    )

    is_empty = snapshot_total_size == 0

    swh_snp_id = str(
        CoreSWHID(object_type=ObjectType.SNAPSHOT, object_id=hash_to_bytes(snapshot_id))
    )

    if visit_info:
        timestamp = format_utc_iso_date(visit_info["date"])

    if origin_info:
        browse_view_name = f"browse-origin-{browse_context}"
    else:
        browse_view_name = f"browse-snapshot-{browse_context}"

    release_id = None
    root_directory = None

    if path is not None:
        query_params["path"] = path

    if snapshot_total_size and revision_id is not None:
        # browse specific revision for a snapshot requested
        revision = archive.lookup_revision(revision_id)
        root_directory = revision["directory"]
        branches.append(
            SnapshotBranchInfo(
                name=revision_id,
                alias=False,
                target_type="revision",
                target=revision_id,
                directory=root_directory,
                date=revision["date"],
                message=revision["message"],
                url=None,
            )
        )
        query_params["revision"] = revision_id
    elif snapshot_total_size and release_name:
        # browse specific release for a snapshot requested
        release = _get_release(releases, release_name, snapshot_id)
        if release is None:
            _branch_not_found(
                "release",
                release_name,
                snapshot_id,
                snapshot_sizes,
                origin_info,
                timestamp,
                visit_id,
            )
        else:
            if release["target_type"] == "revision":
                revision = archive.lookup_revision(release["target"])
                root_directory = revision["directory"]
                revision_id = release["target"]
            elif release["target_type"] == "directory":
                root_directory = release["target"]
            release_id = release["id"]
            query_params["release"] = release_name
    elif snapshot_total_size:
        head = aliases.get("HEAD")
        if branch_name:
            # browse specific branch for a snapshot requested
            query_params["branch"] = branch_name
            branch = _get_branch(branches, branch_name, snapshot_id)
            if branch is None:
                _branch_not_found(
                    "branch",
                    branch_name,
                    snapshot_id,
                    snapshot_sizes,
                    origin_info,
                    timestamp,
                    visit_id,
                )
            else:
                branch_name = branch["name"]
                if branch["target_type"] == "revision":
                    revision_id = branch["target"]
                    root_directory = branch["directory"]
                elif branch["target_type"] == "directory":
                    root_directory = branch["target"]
        elif head is not None:
            # otherwise, browse branch targeted by the HEAD alias if it exists
            if head["target_type"] in ("content", "directory", "revision"):
                branch_name = "HEAD"
                if head["target_type"] == "revision":
                    # HEAD alias targets a revision
                    head_rev = archive.lookup_revision(head["target"])
                    revision_id = head_rev["id"]
                    root_directory = head_rev["directory"]
                elif head["target_type"] == "directory":
                    # HEAD alias targets a directory
                    root_directory = head["target"]
            elif head["target_type"] == "release":
                # HEAD alias targets a release
                release_name = archive.lookup_release(head["target"])["name"]
                head_rel = _get_release(releases, release_name, snapshot_id)
                if head_rel is None:
                    _branch_not_found(
                        "release",
                        str(release_name),
                        snapshot_id,
                        snapshot_sizes,
                        origin_info,
                        timestamp,
                        visit_id,
                    )
                elif head_rel["target_type"] == "revision":
                    revision = archive.lookup_revision(head_rel["target"])
                    root_directory = revision["directory"]
                    revision_id = head_rel["target"]
                elif head_rel["target_type"] == "directory":
                    root_directory = head_rel["target"]
                if head_rel is not None:
                    release_id = head_rel["id"]
        elif branches:
            # fallback to browse first branch otherwise
            branch = branches[0]
            branch_name = branch["name"]
            revision_id = (
                branch["target"] if branch["target_type"] == "revision" else None
            )
            if branch["target_type"] == "revision":
                root_directory = branch["directory"]
            elif branch["target_type"] == "directory":
                root_directory = branch["target"]
        elif releases:
            # fallback to browse latest release otherwise
            # (releases are sorted by descending order of their names)
            release = releases[0]
            if release["target_type"] == "revision":
                revision = archive.lookup_revision(release["target"])
                root_directory = revision["directory"]
                revision_id = release["target"]
            elif release["target_type"] == "directory":
                root_directory = release["target"]
            release_id = release["id"]
            release_name = release["name"]

    for b in branches:
        branch_query_params = dict(query_params)
        branch_query_params.pop("release", None)
        if b["name"] != b["target"]:
            branch_query_params.pop("revision", None)
            branch_query_params["branch"] = b["name"]
        b["url"] = reverse(
            browse_view_name, url_args=url_args, query_params=branch_query_params
        )

    for r in releases:
        release_query_params = dict(query_params)
        release_query_params.pop("branch", None)
        release_query_params.pop("revision", None)
        release_query_params["release"] = r["name"]
        r["url"] = reverse(
            browse_view_name,
            url_args=url_args,
            query_params=release_query_params,
        )

    revision_info = None
    if revision_id:
        try:
            revision_info = archive.lookup_revision(revision_id)
        except NotFoundExc:
            pass
        else:
            revision_info["date"] = format_utc_iso_date(revision_info["date"])
            revision_info["committer_date"] = format_utc_iso_date(
                revision_info["committer_date"]
            )
            if revision_info["message"]:
                message_lines = revision_info["message"].split("\n")
                revision_info["message_header"] = message_lines[0]
            else:
                revision_info["message_header"] = ""

    snapshot_context = SnapshotContext(
        browse_url=browse_url,
        branch=branch_name,
        branch_alias=branch_name in aliases,
        branches=branches,
        branches_url=branches_url,
        is_empty=is_empty,
        origin_info=origin_info,
        origin_visits_url=origin_visits_url,
        release=release_name,
        release_alias=release_name in aliases,
        release_id=release_id,
        query_params=query_params,
        releases=releases,
        releases_url=releases_url,
        revision_id=revision_id,
        revision_info=revision_info,
        root_directory=root_directory,
        snapshot_id=snapshot_id,
        snapshot_sizes=snapshot_sizes,
        snapshot_swhid=swh_snp_id,
        url_args=url_args,
        visit_info=visit_info,
    )

    if revision_info:
        revision_info["revision_url"] = gen_revision_url(
            revision_info["id"], snapshot_context
        )

    return snapshot_context


def _build_breadcrumbs(
    snapshot_context: SnapshotContext, path: Optional[str]
) -> List[Dict[str, str]]:
    origin_info = snapshot_context["origin_info"]
    url_args = snapshot_context["url_args"]
    query_params = dict(snapshot_context["query_params"])
    root_directory = snapshot_context["root_directory"]

    path_info = gen_path_info(path)

    if origin_info:
        browse_view_name = "browse-origin-directory"
    else:
        browse_view_name = "browse-snapshot-directory"

    breadcrumbs = []
    if root_directory:
        query_params.pop("path", None)
        breadcrumbs.append(
            {
                "name": root_directory[:7],
                "url": reverse(
                    browse_view_name, url_args=url_args, query_params=query_params
                ),
            }
        )
    for pi in path_info:
        query_params["path"] = pi["path"]
        breadcrumbs.append(
            {
                "name": pi["name"],
                "url": reverse(
                    browse_view_name, url_args=url_args, query_params=query_params
                ),
            }
        )
    return breadcrumbs


def _check_origin_url(snapshot_id: Optional[str], origin_url: Optional[str]) -> None:
    if snapshot_id is None and origin_url is None:
        raise BadInputExc("An origin URL must be provided as query parameter.")


def browse_snapshot_directory(
    request: HttpRequest,
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
    path: Optional[str] = None,
) -> HttpResponse:
    """
    Django view implementation for browsing a directory in a snapshot context.
    """
    _check_origin_url(snapshot_id, origin_url)

    visit_id = int(request.GET.get("visit_id", 0))
    snapshot_context = get_snapshot_context(
        snapshot_id=snapshot_id,
        origin_url=origin_url,
        timestamp=timestamp,
        visit_id=visit_id or None,
        path=path,
        browse_context="directory",
        branch_name=request.GET.get("branch"),
        release_name=request.GET.get("release"),
        revision_id=request.GET.get("revision"),
        visit_type=request.GET.get("visit_type"),
    )

    root_directory = snapshot_context["root_directory"]

    if root_directory is None and snapshot_context["branch"] is not None:
        branch_info = [
            branch
            for branch in snapshot_context["branches"]
            if branch["name"] == snapshot_context["branch"]
        ]
        # special case where the branch to browse targets a content instead of a directory
        if branch_info and branch_info[0]["target_type"] == "content":
            # redirect to browse content view
            if "origin_url" not in snapshot_context["query_params"]:
                snapshot_id = snapshot_context["snapshot_id"]
                snapshot_context["query_params"]["snapshot"] = snapshot_id
            return redirect(
                reverse(
                    "browse-content",
                    url_args={"query_string": f"sha1_git:{branch_info[0]['target']}"},
                    query_params=snapshot_context["query_params"],
                )
            )

    sha1_git = root_directory
    target_type = None
    error_info: Dict[str, Any] = {
        "status_code": 200,
        "description": None,
    }
    if root_directory and path:
        try:
            dir_info = archive.lookup_directory_with_path(root_directory, path)
            sha1_git = dir_info["target"]
            target_type = dir_info["type"]
        except NotFoundExc as e:
            sha1_git = None
            error_info["status_code"] = 404
            error_info["description"] = f"NotFoundExc: {str(e)}"

    if target_type == "file":
        browse_content_url = reverse(
            "browse-content",
            url_args={"query_string": f"sha1_git:{sha1_git}"},
            query_params=request.GET.dict(),
        )
        return redirect(browse_content_url)

    dirs = []
    files = []
    if sha1_git:
        dirs, files = get_directory_entries(sha1_git)

    origin_info = snapshot_context["origin_info"]
    visit_info = snapshot_context["visit_info"]
    url_args = snapshot_context["url_args"]
    query_params = dict(snapshot_context["query_params"])
    revision_id = snapshot_context["revision_id"]
    snapshot_id = snapshot_context["snapshot_id"]

    if origin_info:
        browse_view_name = "browse-origin-directory"
    else:
        browse_view_name = "browse-snapshot-directory"

    breadcrumbs = _build_breadcrumbs(snapshot_context, path)

    path = "" if path is None else (path + "/")

    for d in dirs:
        if d["type"] == "rev":
            d["url"] = reverse("browse-revision", url_args={"sha1_git": d["target"]})
        else:
            query_params["path"] = path + d["name"]
            d["url"] = reverse(
                browse_view_name, url_args=url_args, query_params=query_params
            )

    sum_file_sizes = 0

    readmes = {}

    if origin_info:
        browse_view_name = "browse-origin-content"
    else:
        browse_view_name = "browse-snapshot-content"

    for f in files:
        query_params["path"] = path + f["name"]
        f["url"] = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )
        if f["length"] is not None:
            sum_file_sizes += f["length"]
        if f["name"].lower().startswith("readme"):
            readmes[f["name"]] = f.get("target")

    readme_name, readme_url, readme_html = get_readme_to_display(readmes)

    if origin_info:
        browse_view_name = "browse-origin-log"
    else:
        browse_view_name = "browse-snapshot-log"

    history_url = None
    if snapshot_id != _empty_snapshot_id:
        query_params.pop("path", None)
        history_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    nb_files = None
    nb_dirs = None
    dir_path = None
    if root_directory:
        nb_files = len(files)
        nb_dirs = len(dirs)
        dir_path = "/" + path

    swh_objects = []
    vault_cooking: Dict[str, Any] = {
        "directory_context": False,
        "directory_swhid": None,
        "revision_context": False,
        "revision_swhid": None,
    }
    revision_found = False

    if revision_id is not None:
        try:
            archive.lookup_revision(revision_id)
        except NotFoundExc:
            pass
        else:
            revision_found = True

    if sha1_git is not None:
        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.DIRECTORY, object_id=sha1_git)
        )
        vault_cooking.update(
            {
                "directory_context": True,
                "directory_swhid": f"swh:1:dir:{sha1_git}",
            }
        )
    if revision_id is not None and revision_found:
        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.REVISION, object_id=revision_id)
        )
        vault_cooking.update(
            {
                "revision_context": True,
                "revision_swhid": f"swh:1:rev:{revision_id}",
            }
        )
    swh_objects.append(
        SWHObjectInfo(object_type=ObjectType.SNAPSHOT, object_id=snapshot_id)
    )

    visit_date = None
    visit_type = None
    if visit_info:
        visit_date = format_utc_iso_date(visit_info["date"])
        visit_type = visit_info["type"]

    release_id = snapshot_context["release_id"]
    if release_id:
        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.RELEASE, object_id=release_id)
        )

    dir_metadata = DirectoryMetadata(
        object_type=ObjectType.DIRECTORY,
        object_id=sha1_git,
        directory=sha1_git,
        nb_files=nb_files,
        nb_dirs=nb_dirs,
        sum_file_sizes=sum_file_sizes,
        root_directory=root_directory,
        path=dir_path,
        revision=revision_id,
        revision_found=revision_found,
        release=release_id,
        snapshot=snapshot_id,
        origin_url=origin_url,
        visit_date=visit_date,
        visit_type=visit_type,
    )

    swhids_info = get_swhids_info(swh_objects, snapshot_context, dir_metadata)

    dir_path = "/".join([bc["name"] for bc in breadcrumbs]) + "/"
    context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
    if origin_info:
        context_found = "origin: %s" % origin_info["url"]
    heading = "Directory - %s - %s - %s" % (
        dir_path,
        snapshot_context["branch"],
        context_found,
    )

    top_right_link = None
    if not snapshot_context["is_empty"] and revision_found:
        top_right_link = {
            "url": history_url,
            "icon": swh_object_icons["revisions history"],
            "text": "History",
        }

    return render(
        request,
        "browse-directory.html",
        {
            "heading": heading,
            "swh_object_name": "Directory",
            "swh_object_metadata": dir_metadata,
            "dirs": dirs,
            "files": files,
            "breadcrumbs": breadcrumbs if root_directory else [],
            "top_right_link": top_right_link,
            "readme_name": readme_name,
            "readme_url": readme_url,
            "readme_html": readme_html,
            "snapshot_context": snapshot_context,
            "vault_cooking": vault_cooking,
            "show_actions": True,
            "swhids_info": swhids_info,
            "error_code": error_info["status_code"],
            "error_message": http_status_code_message.get(error_info["status_code"]),
            "error_description": error_info["description"],
        },
        status=error_info["status_code"],
    )


PER_PAGE = 100


def browse_snapshot_log(
    request: HttpRequest,
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
) -> HttpResponse:
    """
    Django view implementation for browsing a revision history in a
    snapshot context.
    """
    _check_origin_url(snapshot_id, origin_url)

    visit_id = int(request.GET.get("visit_id", 0))
    snapshot_context = get_snapshot_context(
        snapshot_id=snapshot_id,
        origin_url=origin_url,
        timestamp=timestamp,
        visit_id=visit_id or None,
        browse_context="log",
        branch_name=request.GET.get("branch"),
        release_name=request.GET.get("release"),
        revision_id=request.GET.get("revision"),
        visit_type=request.GET.get("visit_type"),
    )

    revision_id = snapshot_context["revision_id"]

    if revision_id is None:
        raise NotFoundExc("No revisions history found in the current snapshot context.")

    per_page = int(request.GET.get("per_page", PER_PAGE))
    offset = int(request.GET.get("offset", 0))
    revs_ordering = request.GET.get("revs_ordering", "committer_date")
    cache_key = f"rev_{revision_id}_log_ordering_{revs_ordering}"
    rev_log_session = cache_get(
        cache_key,
        extra_decoders={
            "rev_walker_state": lambda d: State(
                done=d["done"],
                last_rev=d["last_rev"],
                num_revs=d["num_revs"],
                missing_revs=d["missing_revs"],
                revs_to_visit=(
                    list(map(tuple, d["revs_to_visit"]))
                    if revs_ordering == "committer_date"
                    else d["revs_to_visit"]
                ),
            ),
            "set": set,
            "deque": deque,
        },
    )
    rev_log = []
    revs_walker_state = None
    if rev_log_session:
        rev_log = rev_log_session["rev_log"]
        revs_walker_state = rev_log_session["revs_walker_state"]

    if len(rev_log) < offset + per_page:
        revs_walker = archive.get_revisions_walker(
            revs_ordering,
            revision_id,
            max_revs=offset + per_page + 1,
            state=revs_walker_state,
        )
        rev_log += [rev["id"] for rev in revs_walker]
        revs_walker_state = revs_walker.export_state()
        cache_set(
            cache_key,
            {
                "rev_log": rev_log,
                "revs_walker_state": revs_walker_state,
            },
            timeout=60 * 60,  # one hour
            extra_encoders=[
                (State, "rev_walker_state", dataclasses.asdict),
                (set, "set", list),
                (deque, "deque", list),
            ],
        )

    revs = rev_log[offset : offset + per_page]
    revision_log = archive.lookup_revision_multiple(revs)

    origin_info = snapshot_context["origin_info"]
    visit_info = snapshot_context["visit_info"]
    url_args = snapshot_context["url_args"]
    query_params = snapshot_context["query_params"]
    snapshot_id = snapshot_context["snapshot_id"]

    query_params["per_page"] = str(per_page)
    revs_ordering = request.GET.get("revs_ordering", "")
    if revs_ordering:
        query_params["revs_ordering"] = revs_ordering

    if origin_info:
        browse_view_name = "browse-origin-log"
    else:
        browse_view_name = "browse-snapshot-log"

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        query_params["offset"] = str(offset + per_page)
        prev_log_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    next_log_url = None
    if offset != 0:
        query_params["offset"] = str(offset - per_page)
        next_log_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    revision_log_data = format_log_entries(revision_log, per_page, snapshot_context)

    browse_rev_link = gen_revision_link(revision_id)

    browse_log_link = gen_revision_log_link(revision_id)

    browse_snp_link = gen_snapshot_link(snapshot_id)

    revision_metadata = {
        "context-independent revision": browse_rev_link,
        "context-independent revision history": browse_log_link,
        "context-independent snapshot": browse_snp_link,
        "snapshot": snapshot_id,
        "object_type": "snapshot",
    }

    if origin_info and visit_info:
        revision_metadata["origin url"] = origin_info["url"]
        revision_metadata["origin visit date"] = format_utc_iso_date(visit_info["date"])
        revision_metadata["origin visit type"] = visit_info["type"]

    swh_objects = [
        SWHObjectInfo(object_type=ObjectType.REVISION, object_id=revision_id),
        SWHObjectInfo(object_type=ObjectType.SNAPSHOT, object_id=snapshot_id),
    ]

    release_id = snapshot_context["release_id"]
    if release_id:
        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.RELEASE, object_id=release_id)
        )
        browse_rel_link = gen_release_link(release_id)
        revision_metadata["release"] = release_id
        revision_metadata["context-independent release"] = browse_rel_link

    swhids_info = get_swhids_info(swh_objects, snapshot_context)

    context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
    if origin_info:
        context_found = "origin: %s" % origin_info["url"]
    heading = "Revision history - %s - %s" % (snapshot_context["branch"], context_found)

    return render(
        request,
        "browse-revision-log.html",
        {
            "heading": heading,
            "swh_object_name": "Revisions history",
            "swh_object_metadata": revision_metadata,
            "revision_log": revision_log_data,
            "revs_ordering": revs_ordering,
            "next_log_url": next_log_url,
            "prev_log_url": prev_log_url,
            "breadcrumbs": None,
            "top_right_link": None,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions": True,
            "swhids_info": swhids_info,
        },
    )


def browse_snapshot_branches(
    request: HttpRequest,
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
    branch_name_include: Optional[str] = None,
) -> HttpResponse:
    """
    Django view implementation for browsing a list of branches in a snapshot
    context.
    """
    _check_origin_url(snapshot_id, origin_url)

    visit_id = int(request.GET.get("visit_id", 0))
    snapshot_context = get_snapshot_context(
        snapshot_id=snapshot_id,
        origin_url=origin_url,
        timestamp=timestamp,
        visit_id=visit_id or None,
        visit_type=request.GET.get("visit_type"),
    )

    branches_bc_str = request.GET.get("branches_breadcrumbs", "")
    branches_bc = branches_bc_str.split(",") if branches_bc_str else []
    branches_from = branches_bc[-1] if branches_bc else ""

    origin_info = snapshot_context["origin_info"]
    url_args = snapshot_context["url_args"]
    query_params = snapshot_context["query_params"]

    if origin_info:
        browse_view_name = "browse-origin-directory"
    else:
        browse_view_name = "browse-snapshot-directory"

    snapshot = archive.lookup_snapshot(
        snapshot_context["snapshot_id"],
        branches_from,
        PER_PAGE,
        target_types=["content", "directory", "revision", "alias"],
        branch_name_include_substring=branch_name_include,
    )
    displayed_branches: List[Dict[str, Any]] = []
    if snapshot:
        branches, _, _ = process_snapshot_branches(snapshot)
        displayed_branches = [dict(branch) for branch in branches]

    for branch in displayed_branches:
        query_params = {"snapshot": snapshot_id, "branch": branch["name"]}
        if origin_info:
            query_params["origin_url"] = origin_info["url"]

        directory_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )
        branch["directory_url"] = directory_url
        del query_params["branch"]

        if branch["target_type"] in ("directory", "revision"):
            target_url = reverse(
                f"browse-{branch['target_type']}",
                url_args={"sha1_git": branch["target"]},
                query_params=query_params,
            )
        elif branch["target_type"] == "content":
            target_url = reverse(
                "browse-content",
                url_args={"query_string": f"sha1_git:{branch['target']}"},
                query_params=query_params,
            )
        branch["target_url"] = target_url
        branch["tooltip"] = (
            f"The branch {branch['name']} targets "
            f"{branch['target_type']} {branch['target']}"
        )

    if origin_info:
        browse_view_name = "browse-origin-branches"
    else:
        browse_view_name = "browse-snapshot-branches"

    prev_branches_url = None
    next_branches_url = None

    if branches_bc:
        query_params_prev = dict(query_params)

        query_params_prev["branches_breadcrumbs"] = ",".join(branches_bc[:-1])
        prev_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_prev
        )
    elif branches_from:
        prev_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    if snapshot and snapshot["next_branch"] is not None:
        query_params_next = dict(query_params)
        next_branch = snapshot["next_branch"]
        branches_bc.append(next_branch)
        query_params_next["branches_breadcrumbs"] = ",".join(branches_bc)
        next_branches_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_next
        )

    heading = "Branches - "
    if origin_info:
        heading += "origin: %s" % origin_info["url"]
    else:
        heading += "snapshot: %s" % snapshot_id

    return render(
        request,
        "browse-branches.html",
        {
            "heading": heading,
            "swh_object_name": "Branches",
            "swh_object_metadata": {},
            "top_right_link": None,
            "displayed_branches": displayed_branches,
            "prev_branches_url": prev_branches_url,
            "next_branches_url": next_branches_url,
            "snapshot_context": snapshot_context,
            "search_string": branch_name_include or "",
        },
    )


def browse_snapshot_releases(
    request: HttpRequest,
    snapshot_id: Optional[str] = None,
    origin_url: Optional[str] = None,
    timestamp: Optional[str] = None,
    release_name_include: Optional[str] = None,
):
    """
    Django view implementation for browsing a list of releases in a snapshot
    context.
    """
    _check_origin_url(snapshot_id, origin_url)

    visit_id = int(request.GET.get("visit_id", 0))
    snapshot_context = get_snapshot_context(
        snapshot_id=snapshot_id,
        origin_url=origin_url,
        timestamp=timestamp,
        visit_id=visit_id or None,
        visit_type=request.GET.get("visit_type"),
    )

    rel_bc_str = request.GET.get("releases_breadcrumbs", "")
    rel_bc = rel_bc_str.split(",") if rel_bc_str else []
    rel_from = rel_bc[-1] if rel_bc else ""

    origin_info = snapshot_context["origin_info"]
    url_args = snapshot_context["url_args"]
    query_params = snapshot_context["query_params"]

    snapshot = archive.lookup_snapshot(
        snapshot_context["snapshot_id"],
        rel_from,
        PER_PAGE,
        target_types=["release", "alias"],
        branch_name_include_substring=release_name_include,
    )

    displayed_releases: List[Dict[str, Any]] = []
    if snapshot:
        _, releases, _ = process_snapshot_branches(snapshot)
        displayed_releases = [dict(release) for release in releases]

    for release in displayed_releases:
        query_params_tgt = {"snapshot": snapshot_id, "release": release["name"]}
        if origin_info:
            query_params_tgt["origin_url"] = origin_info["url"]

        release_url = reverse(
            "browse-release",
            url_args={"sha1_git": release["id"]},
            query_params=query_params_tgt,
        )

        target_url = ""
        tooltip = (
            f"The release {release['name']} targets "
            f"{release['target_type']} {release['target']}"
        )
        if release["target_type"] == "revision":
            target_url = reverse(
                "browse-revision",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "directory":
            target_url = reverse(
                "browse-directory",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "content":
            target_url = reverse(
                "browse-content",
                url_args={"query_string": release["target"]},
                query_params=query_params_tgt,
            )
        elif release["target_type"] == "release":
            target_url = reverse(
                "browse-release",
                url_args={"sha1_git": release["target"]},
                query_params=query_params_tgt,
            )
            tooltip = (
                f"The release {release['name']} "
                f"is an alias for release {release['target']}"
            )

        release["release_url"] = release_url
        release["target_url"] = target_url
        release["tooltip"] = tooltip

    if origin_info:
        browse_view_name = "browse-origin-releases"
    else:
        browse_view_name = "browse-snapshot-releases"

    prev_releases_url = None
    next_releases_url = None

    if rel_bc:
        query_params_prev = dict(query_params)

        query_params_prev["releases_breadcrumbs"] = ",".join(rel_bc[:-1])
        prev_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_prev
        )
    elif rel_from:
        prev_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params
        )

    if snapshot and snapshot["next_branch"] is not None:
        query_params_next = dict(query_params)
        next_rel = snapshot["next_branch"]
        rel_bc.append(next_rel)
        query_params_next["releases_breadcrumbs"] = ",".join(rel_bc)
        next_releases_url = reverse(
            browse_view_name, url_args=url_args, query_params=query_params_next
        )

    heading = "Releases - "
    if origin_info:
        heading += "origin: %s" % origin_info["url"]
    else:
        heading += "snapshot: %s" % snapshot_id

    return render(
        request,
        "browse-releases.html",
        {
            "heading": heading,
            "top_panel_visible": False,
            "top_panel_collapsible": False,
            "swh_object_name": "Releases",
            "swh_object_metadata": {},
            "top_right_link": None,
            "displayed_releases": displayed_releases,
            "prev_releases_url": prev_releases_url,
            "next_releases_url": next_releases_url,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions": False,
            "search_string": release_name_include or "",
        },
    )



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pytest_django.asserts import assertContains, assertNotContains, assertTemplateUsed

assert_template_used = assertTemplateUsed
assert_contains = assertContains
assert_not_contains = assertNotContains



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("swh_web_add_forge_now", "0009_canonicalise_forge_urls"),
    ]

    operations = [
        migrations.AlterField(
            model_name="request",
            name="forge_contact_comment",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Where did you find this contact information (url, ...)",
            ),
        ),
    ]



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from importlib.util import find_spec
from typing import List, Union

from django.conf import settings
from django.conf.urls import handler400, handler403, handler404, handler500, include
from django.contrib.staticfiles.views import serve
from django.urls import URLPattern, URLResolver
from django.urls import re_path as url

from swh.web.config import get_config
from swh.web.utils.exc import swh_handle400, swh_handle403, swh_handle404, swh_handle500

swh_web_config = get_config()


urlpatterns: List[Union[URLPattern, URLResolver]] = []

# Register URLs for each SWH Django application
for app in settings.SWH_DJANGO_APPS:
    app_urls = app + ".urls"
    try:
        app_urls_spec = find_spec(app_urls)
        if app_urls_spec is not None:
            urlpatterns.append(url(r"^", include(app_urls)))
    except ModuleNotFoundError:
        assert False, f"Django application {app} not found !"


# allow to serve assets through django staticfiles
# even if settings.DEBUG is False
def insecure_serve(request, path, **kwargs):
    return serve(request, path, insecure=True, **kwargs)


# enable to serve compressed assets through django development server
if swh_web_config["serve_assets"]:
    static_pattern = r"^%s(?P<path>.*)$" % settings.STATIC_URL[1:]
    urlpatterns.append(url(static_pattern, insecure_serve, name="insecure-serve"))


handler400 = swh_handle400  # noqa
handler403 = swh_handle403  # noqa
handler404 = swh_handle404  # noqa
handler500 = swh_handle500  # noqa



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime
from pathlib import Path

from django.templatetags.static import static
from django.utils import timezone

from swh.web import config
from swh.web.tests.django_asserts import assert_contains, assert_not_contains
from swh.web.tests.helpers import check_html_get_response, check_http_get_response
from swh.web.utils import reverse
from swh.web.webapp.urls import SWH_FAVICON


def test_homepage_view(client):
    url = reverse("swh-web-homepage")
    check_html_get_response(client, url, status_code=200, template_used="homepage.html")


def test_homepage_view_no_counters_config(client, mocker):
    updated_config = dict(config.get_config())
    updated_config.pop("counters")
    mocker.patch.object(config, "swhweb_config", updated_config)
    url = reverse("swh-web-homepage")
    resp = check_html_get_response(
        client, url, status_code=200, template_used="homepage.html"
    )
    assert_not_contains(resp, "swh-counter")


def test_homepage_view_no_counters(client, mocker):
    mock_archive = mocker.patch("swh.web.webapp.urls.archive")
    mock_archive.stat_counters.return_value = {}
    url = reverse("swh-web-homepage")
    resp = check_html_get_response(
        client, url, status_code=200, template_used="homepage.html"
    )
    assert_not_contains(resp, "swh-counter")


def test_homepage_view_counters_missing(client, mocker):
    mock_archive = mocker.patch("swh.web.webapp.urls.archive")
    mock_archive.stat_counters.return_value = {
        "content": 150,
        "directory": 45,
        "revision": 78,
    }
    url = reverse("swh-web-homepage")
    resp = check_html_get_response(
        client, url, status_code=200, template_used="homepage.html"
    )
    assert_contains(resp, "swh-content-count")
    assert_contains(resp, "swh-directory-count")
    assert_contains(resp, "swh-revision-count")
    assert_not_contains(resp, "swh-person-count")
    assert_not_contains(resp, "swh-origin-count")
    assert_not_contains(resp, "swh-release-count")


def test_js_reverse_view(client):
    url = reverse("js-reverse")
    check_http_get_response(client, url, status_code=200)


def test_favicon_view(client):
    url = reverse("favicon")
    resp = check_http_get_response(client, url, status_code=301)
    assert resp["location"] == static(SWH_FAVICON)


def test_securitytxt_static(settings):
    security_txt = Path(settings.STATIC_DIR) / "security.txt"
    # static/security.txt contains a static expiration date, this test will fail
    # if it expires in less than 30 days so we can check the file content is still
    # valid
    expiration_date = datetime.fromisoformat(
        security_txt.read_text().splitlines()[1].split(": ")[1]
    )
    delta = expiration_date - timezone.now()
    assert delta.days > 30



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from corsheaders.middleware import ACCESS_CONTROL_ALLOW_ORIGIN

from swh.model.hashutil import hash_to_bytes
from swh.model.swhids import ObjectType, QualifiedSWHID
from swh.web.badges import badge_config, get_logo_data
from swh.web.tests.django_asserts import assert_contains
from swh.web.tests.helpers import check_http_get_response
from swh.web.utils import archive, reverse
from swh.web.utils.identifiers import resolve_swhid


def test_content_badge(client, content):
    _test_badge_endpoints(client, "content", content["sha1_git"])


def test_directory_badge(client, directory):
    _test_badge_endpoints(client, "directory", directory)


def test_origin_badge(client, origin):
    _test_badge_endpoints(client, "origin", origin["url"])


def test_release_badge(client, release):
    _test_badge_endpoints(client, "release", release)


def test_revision_badge(client, revision):
    _test_badge_endpoints(client, "revision", revision)


def test_snapshot_badge(client, snapshot):
    _test_badge_endpoints(client, "snapshot", snapshot)


def test_badge_errors(
    client,
    unknown_content,
    unknown_directory,
    unknown_release,
    unknown_revision,
    unknown_snapshot,
    invalid_sha1,
):
    new_origin = "https://example.org/not/found"
    for object_type, object_id in (
        ("content", unknown_content["sha1_git"]),
        ("directory", unknown_directory),
        ("origin", new_origin),
        ("release", unknown_release),
        ("revision", unknown_revision),
        ("snapshot", unknown_snapshot),
    ):
        url_args = {"object_type": object_type, "object_id": object_id}
        url = reverse("swh-badge", url_args=url_args)
        resp = check_http_get_response(
            client, url, status_code=404, content_type="image/svg+xml"
        )
        _check_generated_badge(resp, **url_args, error="not found", status_code=404)

    for object_type, object_id in (
        (ObjectType.CONTENT, invalid_sha1),
        (ObjectType.DIRECTORY, invalid_sha1),
        (ObjectType.RELEASE, invalid_sha1),
        (ObjectType.REVISION, invalid_sha1),
        (ObjectType.SNAPSHOT, invalid_sha1),
        (ObjectType.SNAPSHOT, "123"),
    ):
        url_args = {"object_type": object_type.name.lower(), "object_id": object_id}
        url = reverse("swh-badge", url_args=url_args)

        resp = check_http_get_response(
            client, url, status_code=400, content_type="image/svg+xml"
        )
        _check_generated_badge(resp, **url_args, error="invalid id", status_code=400)

        object_swhid = f"swh:1:{object_type.value}:{object_id}"
        url = reverse("swh-badge-swhid", url_args={"object_swhid": object_swhid})
        resp = check_http_get_response(
            client, url, status_code=400, content_type="image/svg+xml"
        )
        _check_generated_badge(resp, "", "", error="invalid id", status_code=400)


def test_badge_endpoints_have_cors_header(client, origin, release):
    url = reverse(
        "swh-badge", url_args={"object_type": "origin", "object_id": origin["url"]}
    )

    resp = check_http_get_response(
        client,
        url,
        status_code=200,
        content_type="image/svg+xml",
        http_origin="https://example.org",
    )
    assert ACCESS_CONTROL_ALLOW_ORIGIN in resp

    release_swhid = str(
        QualifiedSWHID(object_type=ObjectType.RELEASE, object_id=hash_to_bytes(release))
    )
    url = reverse("swh-badge-swhid", url_args={"object_swhid": release_swhid})
    resp = check_http_get_response(
        client,
        url,
        status_code=200,
        content_type="image/svg+xml",
        http_origin="https://example.org",
    )
    assert ACCESS_CONTROL_ALLOW_ORIGIN in resp


def _test_badge_endpoints(client, object_type: str, object_id: str):
    url_args = {"object_type": object_type, "object_id": object_id}
    url = reverse("swh-badge", url_args=url_args)
    resp = check_http_get_response(
        client, url, status_code=200, content_type="image/svg+xml"
    )
    _check_generated_badge(resp, **url_args)

    if object_type != "origin":
        obj_swhid = str(
            QualifiedSWHID(
                object_type=ObjectType[object_type.upper()],
                object_id=hash_to_bytes(object_id),
                origin="https://git.example.org/user/project",
            )
        )
        url = reverse("swh-badge-swhid", url_args={"object_swhid": obj_swhid})
        resp = check_http_get_response(
            client, url, status_code=200, content_type="image/svg+xml"
        )
        _check_generated_badge(resp, **url_args)


def _check_generated_badge(
    response, object_type, object_id, error=None, status_code=200
):
    assert response["Content-Type"] == "image/svg+xml"

    if not object_type:
        object_type = "object"

    if object_type == "origin" and error is None:
        link = reverse("browse-origin", query_params={"origin_url": object_id})
        text = "repository"
        assert "Cache-Control" in response
    elif error is None:
        text = str(
            QualifiedSWHID(
                object_type=ObjectType[object_type.upper()],
                object_id=hash_to_bytes(object_id),
            )
        )
        link = resolve_swhid(text)["browse_url"]
        if object_type == "release":
            release = archive.lookup_release(object_id)
            text = release["name"]
        assert "Cache-Control" in response
    elif error == "invalid id":
        text = "error"
        link = f"invalid {object_type} id"
        object_type = "error"
        assert "Cache-Control" not in response
    elif error == "not found":
        text = "error"
        link = f"{object_type} not found"
        object_type = "error"
        assert "Cache-Control" not in response

    assert_contains(response, "<svg ", status_code=status_code)
    assert_contains(response, "</svg>", status_code=status_code)
    assert_contains(response, get_logo_data(), status_code=status_code)
    assert_contains(
        response, badge_config[object_type]["color"], status_code=status_code
    )
    assert_contains(
        response, badge_config[object_type]["title"], status_code=status_code
    )
    assert_contains(response, text, status_code=status_code)
    assert_contains(response, link, status_code=status_code)



# Copyright (C) 2022-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.http import HttpResponse
from rest_framework.request import Request

from swh.model import model
from swh.model.git_objects import (
    content_git_object,
    directory_git_object,
    release_git_object,
    revision_git_object,
    snapshot_git_object,
)
from swh.model.swhids import ObjectType
from swh.storage.algos.directory import directory_get
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.web import config
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.utils import archive
from swh.web.utils.exc import NotFoundExc
from swh.web.utils.identifiers import parse_core_swhid


@api_route(
    "/raw/<swhid:swhid>/",
    "api-1-raw-object",
    throttle_scope="swh_raw_object",
)
@api_doc("/raw/", category="Archive")
@format_docstring()
def api_raw_object(request: Request, swhid: str):
    """
    .. http:get::  /api/1/raw/(swhid)/

        Get the object corresponding to the SWHID in raw form.

        This endpoint exposes the internal representation (see the
        ``*_git_object`` functions in :mod:`swh.model.git_objects`), and
        so can be used to fetch a binary blob which hashes to the same
        identifier.

        :param string swhid: the object's SWHID

        :resheader Content-Type: application/octet-stream

        :statuscode 200: no error
        :statuscode 404: the requested object cannot be found in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`raw/swh:1:snp:6a3a2cf0b2b90ce7ae1cf0a221ed68035b686f5a`
    """
    parsed_swhid = parse_core_swhid(swhid)
    object_id = parsed_swhid.object_id
    object_type = parsed_swhid.object_type

    storage = config.storage()

    def not_found():
        return NotFoundExc(f"Object with id {swhid} not found.")

    if object_type == ObjectType.CONTENT:
        # fetching content metadata
        try:
            cnt_dict = archive.lookup_content(
                f"sha1_git:{object_id.hex()}", json_convert=False, with_data=True
            )
        except NotFoundExc:
            raise not_found()
        cnt = model.Content.from_dict(cnt_dict)
        result = content_git_object(cnt)

    elif object_type == ObjectType.DIRECTORY:
        dir_ = directory_get(storage, object_id)
        if dir_ is None:
            raise not_found()
        result = directory_git_object(dir_)

    elif object_type == ObjectType.REVISION:
        rev = storage.revision_get([object_id])[0]
        if rev is None:
            raise not_found()
        result = revision_git_object(rev)

    elif object_type == ObjectType.RELEASE:
        rel = storage.release_get([object_id])[0]
        if rel is None:
            raise not_found()
        result = release_git_object(rel)

    elif object_type == ObjectType.SNAPSHOT:
        snp = snapshot_get_all_branches(storage, object_id)
        if snp is None:
            raise not_found()
        result = snapshot_git_object(snp)

    else:
        raise ValueError(f"Unexpected object type variant: {object_type}")

    response = HttpResponse(result, content_type="application/octet-stream")
    filename = swhid.replace(":", "_") + "_raw"
    response["Content-disposition"] = f"attachment; filename={filename}"

    return response



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from swh.web.config import get_config
from swh.web.save_code_now.origin_save import (
    get_save_origin_requests_to_update,
    schedule_origins_recurrent_visits,
    update_save_origin_requests_from_queryset,
)
from swh.web.utils.exc import BadInputExc

try:
    from swh.webhooks.utils import get_verified_webhook_payload

    webhooks_available = True
except ImportError:
    webhooks_available = False


@require_POST
@csrf_exempt
def save_origin_visit_webhook(request):
    """Endpoint that receives webhook messages about origin visits.
    Its purpose is to update the status of a save code now request."""
    if not request.headers.get("X-Swh-Event", "") == "origin.visit":
        raise BadInputExc("POST request is not a Software Heritage webhook")

    try:
        payload = get_verified_webhook_payload(
            request_data=request.body,
            request_headers=request.headers,
            secret=get_config().get("save_code_now_webhook_secret", ""),
        )
    except ValueError:
        raise BadInputExc("Webhook body verification failed")

    origin_url = payload["origin_url"]
    save_requests = get_save_origin_requests_to_update(origin_url=origin_url)

    if save_requests.count() > 0:
        save_requests_info = update_save_origin_requests_from_queryset(save_requests)
        scheduled = schedule_origins_recurrent_visits(save_requests_info)
        message = f"Status of Save Code Now request updated for origin {origin_url}."
        if scheduled:
            message += "\nOrigin was also scheduled for recurrent visits."
    else:
        message = f"No Save Code Now request to update for origin {origin_url}"

    return HttpResponse(message)



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from rest_framework.request import Request

from swh.provenance import get_provenance
from swh.provenance.interface import ProvenanceInterface
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import APIUrls, api_route
from swh.web.auth.utils import API_PROVENANCE_PERMISSION
from swh.web.config import get_config
from swh.web.utils.exc import BadInputExc, ForbiddenExc, UnauthorizedExc
from swh.web.utils.identifiers import parse_core_swhid
from swh.web.utils.url_path_converters import register_url_path_converters

provenance_api_urls = APIUrls()

register_url_path_converters()


def _provenance() -> ProvenanceInterface:
    return get_provenance(**get_config()["provenance"])


def _check_auth_and_permission(request: Request):
    if not bool(request.user and request.user.is_authenticated):
        raise UnauthorizedExc("This API endpoint requires authentication.")
    if (
        not request.user.has_perm(API_PROVENANCE_PERMISSION)
        and not request.user.is_staff
    ):
        raise ForbiddenExc("This API endpoint requires a special user permission.")


@api_route(
    "/provenance/whereis/<swhid:target>/",
    "api-1-provenance-whereis",
    api_urls=provenance_api_urls,
)
@api_doc("/provenance/whereis/", category="Provenance")
@format_docstring()
def api_provenance_whereis(request: Request, target: str):
    """
    .. http:get:: /api/1/provenance/whereis/(target)/

        Given a core SWHID return a qualified SWHID with some provenance info:

        - the release or revision containing that content or directory
        - the url of the origin containing that content or directory

        This can also be called for revision, release or snapshot to retrieve
        origin url information if any. When using a revision, the anchor will
        be an associated release if any.

        .. note::

            The quality of the result is not guaranteed whatsoever. Since the
            definition of "best" likely varies from one usage to the next, this API
            will evolve in the future when this notion gets better defined.

        .. warning::

            That endpoint is not publicly available and requires authentication and
            special user permission in order to request it.

        :param string target: a core SWHID targeting an archived object

        The response is a string containing a qualified SWHID with provenance info.

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: provided core SWHID is invalid
        :statuscode 401: request is not authenticated
        :statuscode 403: user does not have permission to query the endpoint

        **Example:**

        .. parsed-literal::

            :swh_web_api:`provenance/whereis/swh:1:cnt:dcb2d732994e615aab0777bfe625bd1f07e486ac/`
    """
    _check_auth_and_permission(request)
    core_swhid = parse_core_swhid(target)
    provenance = _provenance().whereis(swhid=core_swhid)
    return str(provenance) if provenance is not None else None


@api_route(
    "/provenance/whereare/",
    "api-1-provenance-whereare",
    api_urls=provenance_api_urls,
    methods=["POST"],
)
@api_doc("/provenance/whereare/", category="Provenance")
@format_docstring()
def api_provenance_whereare(request: Request):
    """
    .. http:post:: /api/1/provenance/whereare/

        Given a list of core SWHIDs return qualified SWHIDs with some provenance info.

        See :http:get:`/api/1/provenance/whereis/(target)/` documentation for more details.

        .. warning::

            That endpoint is not publicly available and requires authentication and
            special user permission in order to request it.

        :<jsonarr string -: input array of core SWHIDs

        The response is a JSON array of strings containing qualified SWHIDs with
        provenance info.

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: provided core SWHID is invalid
        :statuscode 401: request is not authenticated
        :statuscode 403: user does not have permission to query the endpoint
    """
    _check_auth_and_permission(request)
    if not request.data or not isinstance(request.data, list):
        raise BadInputExc("A list of core SWHIDS must be provided in POST data.")
    swhids = [parse_core_swhid(swhid) for swhid in request.data]
    provenances = _provenance().whereare(swhids=swhids)
    return [str(provenance) if provenance else None for provenance in provenances]



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from django.urls import get_resolver


def test_swh_web_urls_have_trailing_slash():
    urls = set(
        value[1]
        for key, value in get_resolver().reverse_dict.items()
        if type(key) is str
        and key not in ("browse-swhid", "favicon-no-trailing-slash")
        and key
        and not key.startswith("browse-directory-get-content-at-path")
    )
    for url in urls:
        if url != "$":
            assert url.endswith(("/$", "/\\Z"))


def test_urls_registration_error_for_not_found_django_app(django_settings):
    app_name = "swh.web.foobar"
    with pytest.raises(
        AssertionError, match=f"Django application {app_name} not found !"
    ):
        django_settings.SWH_DJANGO_APPS = django_settings.SWH_DJANGO_APPS + [app_name]



# Copyright (C) 2023-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import uuid

import pytest

from swh.model.hashutil import hash_to_bytes
from swh.model.model import Origin, OriginVisit, OriginVisitStatus
from swh.storage.utils import now
from swh.web.save_code_now.models import (
    SAVE_TASK_PENDING,
    SAVE_TASK_RUNNING,
    SAVE_TASK_SUCCEEDED,
    SaveAuthorizedOrigin,
    SaveOriginRequest,
)
from swh.web.save_code_now.origin_visit_webhook import webhooks_available
from swh.web.tests.helpers import check_api_post_responses, check_http_post_response
from swh.web.utils import reverse
from swh.web.utils.typing import OriginExistenceCheckInfo

pytestmark = [
    pytest.mark.django_db,
    pytest.mark.skipif(not webhooks_available, reason="swh-webhooks is not available"),
]


@pytest.fixture(autouse=True)
def populated_db():
    SaveAuthorizedOrigin.objects.create(url="https://git.example.org/")


def _send_origin_visit_webhook(
    client,
    mocker,
    payload,
    add_webhook_headers=True,
    invalid_signature=False,
    status_code=200,
):
    from swh.webhooks.utils import sign_webhook_payload

    webhook_url = reverse("origin-save-visit-webhook")
    webhook_timestamp = now()
    webhook_secret = "whsec_" + "c" * 32
    webhook_msgid = "some-id"
    if invalid_signature:
        webhook_signature = "foo"
    else:
        webhook_signature = sign_webhook_payload(
            payload=json.dumps(payload),
            timestamp=webhook_timestamp,
            msg_id=webhook_msgid,
            secret=webhook_secret,
        )

    mocker.patch(
        "swh.web.save_code_now.origin_visit_webhook.get_config"
    ).return_value = {"save_code_now_webhook_secret": webhook_secret}

    webhook_headers = {}
    if add_webhook_headers:
        webhook_headers = {
            "HTTP_WEBHOOK_ID": webhook_msgid,
            "HTTP_WEBHOOK_TIMESTAMP": str(int(webhook_timestamp.timestamp())),
            "HTTP_WEBHOOK_SIGNATURE": webhook_signature,
        }

    return check_http_post_response(
        client,
        webhook_url,
        status_code=status_code,
        data=payload,
        HTTP_X_SWH_EVENT="origin.visit",
        **webhook_headers,
    )


def test_save_origin_visit_webhook_invalid_request(client):
    webhook_url = reverse("origin-save-visit-webhook")
    resp = check_http_post_response(client, webhook_url, status_code=400, data={})
    assert b"POST request is not a Software Heritage webhook" in resp.content


def test_save_origin_visit_webhook_missing_webhook_headers(client, mocker):
    resp = _send_origin_visit_webhook(
        client,
        mocker,
        status_code=400,
        payload={"origin_url": "https://git.example.org/user/project"},
        add_webhook_headers=False,
    )
    assert b"Webhook body verification" in resp.content


def test_save_origin_visit_webhook_invalid_signature(client, mocker):
    resp = _send_origin_visit_webhook(
        client,
        mocker,
        status_code=400,
        payload={"origin_url": "https://git.example.org/user/project"},
        invalid_signature=True,
    )
    assert b"Webhook body verification" in resp.content


def test_save_origin_visit_webhook_noop(client, mocker):
    origin_url = "https://git.example.org/user/project"
    resp = _send_origin_visit_webhook(
        client,
        mocker,
        payload={"origin_url": origin_url},
    )
    assert (
        resp.content.decode()
        == f"No Save Code Now request to update for origin {origin_url}"
    )


def test_save_origin_visit_webhook(
    api_client, client, archive_data, mocker, swh_scheduler, snapshot
):
    origin_url = "https://git.example.org/user/project"
    mock_origin_exists = mocker.patch("swh.web.save_code_now.origin_save.origin_exists")
    mock_origin_exists.return_value = OriginExistenceCheckInfo(
        origin_url=origin_url, exists=True, last_modified=None, content_length=None
    )

    # create save request
    url = reverse(
        "api-1-save-origin", url_args={"visit_type": "git", "origin_url": origin_url}
    )
    resp = check_api_post_responses(api_client, url, status_code=200)

    save_request = SaveOriginRequest.objects.first()
    assert save_request.loading_task_status == SAVE_TASK_PENDING
    assert save_request.snapshot_swhid == ""

    # simulate loading task scheduling and execution
    task_id = resp.data["loading_task_id"]
    backend_id = str(uuid.uuid4())
    swh_scheduler.schedule_task_run(task_id, backend_id)
    swh_scheduler.start_task_run(backend_id)

    archive_data.origin_add([Origin(url=origin_url)])
    date = now()
    visit = OriginVisit(origin=origin_url, date=date, type="git")
    visit = archive_data.origin_visit_add([visit])[0]
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=date,
        status="ongoing",
        snapshot=None,
    )
    archive_data.origin_visit_status_add([visit_status])

    # send webhook notifying origin visit is ongoing
    webhook_payload = {
        "origin_url": origin_url,
        "visit_type": "git",
        "visit_date": date.isoformat(),
        "visit_status": "ongoing",
        "snapshot_swhid": None,
    }

    assert (
        _send_origin_visit_webhook(client, mocker, webhook_payload).content.decode()
        == f"Status of Save Code Now request updated for origin {origin_url}."
    )

    # check save request status was updated
    save_request = SaveOriginRequest.objects.first()
    assert save_request.loading_task_status == SAVE_TASK_RUNNING
    assert save_request.snapshot_swhid == ""

    # simulate end of loading task execution
    swh_scheduler.end_task_run(backend_id, "eventful")
    date = now()
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=date,
        status="full",
        snapshot=hash_to_bytes(snapshot),
    )
    archive_data.origin_visit_status_add([visit_status])

    # send webhook notifying origin visit was successful
    snapshot_swhid = f"swh:1:snp:{snapshot}"
    webhook_payload = {
        "origin_url": origin_url,
        "visit_type": "git",
        "visit_date": date.isoformat(),
        "visit_status": "full",
        "snapshot_swhid": snapshot_swhid,
    }

    assert _send_origin_visit_webhook(
        client, mocker, webhook_payload
    ).content.decode() == (
        f"Status of Save Code Now request updated for origin {origin_url}.\n"
        "Origin was also scheduled for recurrent visits."
    )

    # check save request status was updated
    save_request = SaveOriginRequest.objects.first()
    assert save_request.loading_task_status == SAVE_TASK_SUCCEEDED
    assert save_request.snapshot_swhid == snapshot_swhid

    # check next webhook processing is a noop
    assert (
        _send_origin_visit_webhook(client, mocker, webhook_payload).content.decode()
        == f"No Save Code Now request to update for origin {origin_url}"
    )



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.web.add_forge_now.models import Request, RequestStatus


@pytest.mark.parametrize(
    "current_status, allowed_next_statuses",
    [
        (
            RequestStatus.PENDING,
            [
                RequestStatus.WAITING_FOR_FEEDBACK,
                RequestStatus.REJECTED,
                RequestStatus.SUSPENDED,
            ],
        ),
        (
            RequestStatus.WAITING_FOR_FEEDBACK,
            [
                RequestStatus.FEEDBACK_TO_HANDLE,
                RequestStatus.ACCEPTED,
                RequestStatus.REJECTED,
                RequestStatus.SUSPENDED,
                RequestStatus.UNSUCCESSFUL,
            ],
        ),
        (
            RequestStatus.FEEDBACK_TO_HANDLE,
            [
                RequestStatus.WAITING_FOR_FEEDBACK,
                RequestStatus.ACCEPTED,
                RequestStatus.REJECTED,
                RequestStatus.SUSPENDED,
                RequestStatus.UNSUCCESSFUL,
            ],
        ),
        (
            RequestStatus.ACCEPTED,
            [
                RequestStatus.SCHEDULED,
                RequestStatus.REJECTED,
                RequestStatus.UNSUCCESSFUL,
            ],
        ),
        (
            RequestStatus.UNSUCCESSFUL,
            [RequestStatus.ACCEPTED],
        ),
    ],
)
def test_allowed_next_statuses(current_status, allowed_next_statuses):
    assert current_status.allowed_next_statuses() == allowed_next_statuses


@pytest.mark.parametrize(
    "forge_url, expected_domain",
    [
        ("https://gitlab.example.com/foo/bar", "gitlab.example.com"),
        ("gitlab.example.com", "gitlab.example.com"),
        ("gitlab.example.com/foo/bar", "gitlab.example.com"),
    ],
)
def test_request_forge_domain(forge_url, expected_domain):
    assert Request(forge_url=forge_url).forge_domain == expected_domain



# Copyright (C) 2018-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rest_framework.request import Request

from swh.model.hashutil import hash_to_hex
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.utils import archive
from swh.web.utils.exc import LargePayloadExc
from swh.web.utils.identifiers import parse_core_swhid, resolve_swhid


@api_route(r"/resolve/(?P<swhid>.+)/", "api-1-resolve-swhid")
@api_doc("/resolve/", category="Archive")
@format_docstring()
def api_resolve_swhid(request: Request, swhid: str):
    """
    .. http:get:: /api/1/resolve/(swhid)/

        Resolve a SoftWare Hash IDentifier (SWHID)

        Try to resolve a provided `SoftWare Hash IDentifier
        <https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html>`_
        into an url for browsing the pointed archive object.

        If the provided identifier is valid, the existence of the object in
        the archive will also be checked.

        :param string swhid: a SoftWare Hash IDentifier

        :>json string browse_url: the url for browsing the pointed object
        :>json object metadata: object holding optional parts of the SWHID
        :>json string namespace: the SWHID namespace
        :>json string object_id: the hash identifier of the pointed object
        :>json string object_type: the type of the pointed object
        :>json number scheme_version: the scheme version of the SWHID

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid SWHID has been provided
        :statuscode 404: the pointed object does not exist in the archive

        **Example:**

        .. parsed-literal::

            :swh_web_api:`resolve/swh:1:rev:96db9023b881d7cd9f379b0c154650d6c108e9a3;origin=https://github.com/openssl/openssl/`
    """
    # try to resolve the provided swhid
    swhid_resolved = resolve_swhid(swhid)
    # id is well-formed, now check that the pointed
    # object is present in the archive, NotFoundExc
    # will be raised otherwise
    swhid_parsed = swhid_resolved["swhid_parsed"]
    object_type = swhid_parsed.object_type
    object_id = hash_to_hex(swhid_parsed.object_id)
    archive.lookup_object(swhid_parsed.object_type, object_id)
    # id is well-formed and the pointed object exists
    return {
        "namespace": swhid_parsed.namespace,
        "scheme_version": swhid_parsed.scheme_version,
        "object_type": object_type.name.lower(),
        "object_id": object_id,
        "metadata": swhid_parsed.qualifiers(),
        "browse_url": request.build_absolute_uri(swhid_resolved["browse_url"]),
    }


@api_route(r"/known/", "api-1-known", methods=["POST"])
@api_doc("/known/", category="Archive")
@format_docstring()
def api_swhid_known(request: Request):
    """
    .. http:post:: /api/1/known/

        Check if a list of objects are present in the Software Heritage
        archive.

        The objects to check existence must be provided using
        `SoftWare Hash IDentifiers
        <https://docs.softwareheritage.org/devel/swh-model/persistent-identifiers.html>`_.

        :<jsonarr string -: input array of SWHIDs, its length cannot exceed 1000.

        :>json object <swhid>: an object whose keys are input SWHIDs and values
            objects with the following keys:

            * **known (bool)**: whether the object was found

        {common_headers}

        :statuscode 200: no error
        :statuscode 400: an invalid SWHID was provided
        :statuscode 413: the input array of SWHIDs is too large

    """
    limit = 1000
    if len(request.data) > limit:
        raise LargePayloadExc(
            "The maximum number of SWHIDs this endpoint can receive is %s" % limit
        )

    swhids = [parse_core_swhid(swhid) for swhid in request.data]
    known_swhids = archive.lookup_known_swhids(swhids)

    return {swhid: {"known": swhid in known_swhids} for swhid in map(str, swhids)}



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="SaveBulkRequest",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("request_date", models.DateTimeField(auto_now_add=True)),
                ("user_id", models.CharField(max_length=50)),
            ],
            options={
                "db_table": "save_bulk_request",
                "ordering": ["-id"],
            },
        ),
        migrations.CreateModel(
            name="SaveBulkOrigin",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("origin_url", models.CharField(max_length=200)),
                ("visit_type", models.CharField(max_length=30)),
                (
                    "requests",
                    models.ManyToManyField(to="swh_web_save_bulk.savebulkrequest"),
                ),
            ],
            options={
                "db_table": "save_bulk_origin",
                "ordering": ["-id"],
                "indexes": [
                    models.Index(
                        fields=["origin_url", "visit_type"],
                        name="save_bulk_o_origin__69154d_idx",
                    )
                ],
            },
        ),
        migrations.AddConstraint(
            model_name="savebulkorigin",
            constraint=models.UniqueConstraint(
                fields=("origin_url", "visit_type"), name="unicity"
            ),
        ),
    ]



# Copyright (C) 2024-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.conf import settings
from django.contrib import admin
from django.urls import path as url

from swh.web.config import get_config, oidc_enabled

config = get_config()

if oidc_enabled():
    # use swh-auth views in admin site for login/logout when webapp
    # uses Keycloak for authentication
    from swh.auth.django.views import oidc_login, oidc_logout

    admin.site.login = oidc_login  # type: ignore[assignment]
    admin.site.logout = oidc_logout  # type: ignore[assignment]

if "swh.web.add_forge_now" in settings.SWH_DJANGO_APPS:
    # register add forge now request model as manageable by admin site
    from swh.web.add_forge_now.models import Request

    if not admin.site.is_registered(Request):
        admin.site.register(Request)

if "swh.web.save_code_now" in settings.SWH_DJANGO_APPS:
    # register save code now request model as manageable by admin site
    from swh.web.save_code_now.models import SaveOriginRequest

    if not admin.site.is_registered(SaveOriginRequest):
        admin.site.register(SaveOriginRequest)

admin.site.site_header = "swh-web management"
admin.site.site_title = "Software Heritage Web Application management"

admin_urls = url("manage/", admin.site.urls)
# to prevent noisy tracebacks in debug mode when a redirection is performed
setattr(admin_urls, "name", "manage")

urlpatterns = [admin_urls]



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.urls import re_path as url

from swh.web.mailmap.views import (
    admin_mailmap,
    profile_add_mailmap,
    profile_list_mailmap,
    profile_list_mailmap_datatables,
    profile_update_mailmap,
)

urlpatterns = [
    url(
        r"^profile/mailmap/list/$",
        profile_list_mailmap,
        name="profile-mailmap-list",
    ),
    url(
        r"^profile/mailmap/add/$",
        profile_add_mailmap,
        name="profile-mailmap-add",
    ),
    url(
        r"^profile/mailmap/update/$",
        profile_update_mailmap,
        name="profile-mailmap-update",
    ),
    url(
        r"^profile/mailmap/list/datatables/$",
        profile_list_mailmap_datatables,
        name="profile-mailmap-list-datatables",
    ),
    url(r"^admin/mailmap/$", admin_mailmap, name="admin-mailmap"),
]



# Copyright (C) 2022-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.shortcuts import render

from swh.web.add_forge_now.models import RequestStatus
from swh.web.auth.utils import (
    ADD_FORGE_NOW_CHANGE_REQUEST_PERMISSION,
    ADD_FORGE_NOW_VIEW_REQUEST_PERMISSION,
    is_add_forge_now_moderator,
)
from swh.web.utils import reverse

if TYPE_CHECKING:
    from django.http.request import HttpRequest
    from django.http.response import HttpResponse


@user_passes_test(is_add_forge_now_moderator)
def add_forge_now_requests_moderation_dashboard(request: HttpRequest) -> HttpResponse:
    """Moderation dashboard to allow listing current requests."""
    return render(
        request,
        "add-forge-requests-moderation.html",
        {"heading": "Add forge now requests moderation"},
    )


@user_passes_test(is_add_forge_now_moderator)
def add_forge_now_request_dashboard(
    request: HttpRequest, request_id: int
) -> HttpResponse:
    """Moderation dashboard to allow listing current requests."""
    request_edit_url = ""
    if "swh.web.admin" in settings.SWH_DJANGO_APPS and request.user.has_perms(
        (
            ADD_FORGE_NOW_VIEW_REQUEST_PERMISSION,
            ADD_FORGE_NOW_CHANGE_REQUEST_PERMISSION,
        )
    ):
        request_edit_url = reverse(
            "admin:swh_web_add_forge_now_request_change",
            url_args={"object_id": request_id},
        )

    return render(
        request,
        "add-forge-request-dashboard.html",
        {
            "request_id": request_id,
            "heading": "Add forge now request dashboard",
            "next_statuses_for": RequestStatus.next_statuses_str(),
            "request_edit_url": request_edit_url,
        },
    )



# Copyright (C) 2015-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.web.utils.citation import get_bibtex_from_origin, get_bibtex_from_swhid


def test_get_bibtex_from_origin(origin_with_metadata_file):
    citation = get_bibtex_from_origin(origin_with_metadata_file["url"])
    bibtex = citation["content"]

    assert "@software" in bibtex
    assert 'title = "Test Software"' in bibtex


def test_get_bibtex_from_swhid(objects_with_metadata_file):
    citation = get_bibtex_from_swhid(str(objects_with_metadata_file[0]))
    bibtex = citation["content"]

    assert "@software" in bibtex
    assert 'title = "Test Software"' in bibtex
    assert (
        'swhid = "swh:1:snp:89cb4bc962dd297a70053c9288f98d3fde8db740;'
        r'origin=https://git.example.org/repo\_with\_metadata\_file"' in bibtex
    )


def test_get_bibtex_from_codemeta_with_unknown_context(
    origin_url_with_unknown_codemeta_context,
):
    citation = get_bibtex_from_origin(origin_url_with_unknown_codemeta_context)

    assert citation["content"].startswith("@softwareversion{swh-dir-")
    assert (
        "Unknown context URL: https://example.org/codemeta/3.0" in citation["warning"]
    )



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import subprocess

pytest_command = ["python3", "-m", "pytest", "-s"]


def test_random_fixtures():
    """Check random fixture values will be different when random seed
    is not explicitly provided.
    """
    result_first = subprocess.run(
        [*pytest_command, "random_fixtures_test.py"],
        capture_output=True,
        cwd=os.path.dirname(__file__),
    )
    result_second = subprocess.run(
        [*pytest_command, "random_fixtures_test.py"],
        capture_output=True,
        cwd=os.path.dirname(__file__),
    )
    assert result_first.stderr != result_second.stderr
    assert b'Use "pytest --swh-web-random-seed=' in result_first.stdout


def test_random_fixtures_with_seed():
    """Check random fixture values will be the same when random seed
    is explicitly provided through a custom pytest option.
    """
    result_first = subprocess.run(
        [*pytest_command, "--swh-web-random-seed=2021", "random_fixtures_test.py"],
        capture_output=True,
        cwd=os.path.dirname(__file__),
    )
    result_second = subprocess.run(
        [*pytest_command, "--swh-web-random-seed=2021", "random_fixtures_test.py"],
        capture_output=True,
        cwd=os.path.dirname(__file__),
    )
    assert result_first.stderr == result_second.stderr
    assert b'Use "pytest --swh-web-random-seed=2021' in result_first.stdout



# Copyright (C) 2018-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
from datetime import timedelta
import functools
from importlib import import_module, reload
import os
import random
import sys
import time
from typing import Any, Dict, List, Optional

from _pytest.python import Function
from hypothesis import HealthCheck
from hypothesis import settings as hypothesis_settings
import pytest
from pytest_django.fixtures import SettingsWrapper

from django.conf import settings
from django.contrib.auth.models import User
from django.contrib.staticfiles import finders
from django.core.cache import cache
from django.template import engines
from django.template.utils import get_app_template_dirs
from django.test.utils import setup_databases
from django.urls import clear_url_caches
from rest_framework.test import APIClient, APIRequestFactory

from swh.model.hashutil import (
    ALGORITHMS,
    DEFAULT_ALGORITHMS,
    hash_to_bytes,
    hash_to_hex,
)
from swh.model.model import Content, Directory, Revision
from swh.model.swhids import CoreSWHID, ObjectType
from swh.objstorage.interface import objid_from_dict
from swh.scheduler.tests.common import TASK_TYPES
from swh.storage.algos.revisions_walker import get_revisions_walker
from swh.storage.algos.snapshot import (
    snapshot_get_all_branches,
    snapshot_get_latest,
    snapshot_resolve_alias,
)
from swh.web import config as swhweb_config
from swh.web.auth.utils import (
    ADD_FORGE_MODERATOR_PERMISSION,
    ALTER_ADMIN_PERMISSION,
    API_PROVENANCE_PERMISSION,
    API_SAVE_BULK_PERMISSION,
    MAILMAP_ADMIN_PERMISSION,
    MAILMAP_PERMISSION,
    get_or_create_django_permission,
)
from swh.web.config import get_config
from swh.web.save_code_now.origin_save import get_scheduler_load_task_types
from swh.web.settings import common as common_settings
from swh.web.settings import tests as tests_settings
from swh.web.tests.data import (
    ORIGIN_WITH_CFF_FILE,
    ORIGIN_WITH_METADATA_FILES,
    ORIGIN_WITH_QUOTED_SPACE_IN_URL,
    ORIGIN_WITH_UNKNOWN_CODEMETA_CONTEXT,
    get_tests_data,
    override_storages,
    random_content,
    random_sha1,
    random_sha1_bytes,
    random_sha256,
)
from swh.web.utils import browsers_supported_image_mimes, converters
from swh.web.utils.typing import OriginVisitInfo

os.environ["LC_ALL"] = "C.UTF-8"

# Register some hypothesis profiles
hypothesis_settings.register_profile("default", hypothesis_settings())

# we use getattr here to keep mypy happy regardless hypothesis version
function_scoped_fixture_check = (
    [getattr(HealthCheck, "function_scoped_fixture")]
    if hasattr(HealthCheck, "function_scoped_fixture")
    else []
)

suppress_health_check = [
    HealthCheck.too_slow,
    HealthCheck.filter_too_much,
] + function_scoped_fixture_check


hypothesis_settings.register_profile(
    "swh-web",
    hypothesis_settings(
        deadline=None,
        suppress_health_check=suppress_health_check,
    ),
)

hypothesis_settings.register_profile(
    "swh-web-fast",
    hypothesis_settings(
        deadline=None,
        max_examples=5,
        suppress_health_check=suppress_health_check,
    ),
)


def pytest_addoption(parser):
    parser.addoption("--swh-web-random-seed", action="store", default=None)


def pytest_configure(config):
    # Use fast hypothesis profile by default if none has been
    # explicitly specified in pytest option
    if config.getoption("--hypothesis-profile") is None:
        hypothesis_settings.load_profile("swh-web-fast")


_swh_web_custom_section = "swh-web custom section"
_random_seed_cache_key = "swh-web/random-seed"


@pytest.fixture(scope="session", autouse=True)
def webpack_stats_mock(session_mocker):
    """Small hack in order to be able to run the unit tests without static assets
    generated by webpack.

    Those assets are not really needed for the Python tests but the django templates
    will fail to load due to missing generated file webpack-stats.json describing the
    js and css files to include.

    Mock the function loading the webpack-stats.json to overcome that issue.
    """
    test_dir = os.path.dirname(__file__)
    django_apps_dir = os.path.join(test_dir, "../../swh/web")
    bundles = []
    _, apps, _ = next(os.walk(django_apps_dir))
    for app in apps:
        app_assets_dir = os.path.join(django_apps_dir, app, "assets")
        if os.path.exists(app_assets_dir):
            if os.path.exists(os.path.join(app_assets_dir, "index.js")):
                bundles.append(app)
            else:
                _, app_bundles, _ = next(os.walk(app_assets_dir))
                for app_bundle in app_bundles:
                    if os.path.exists(
                        os.path.join(app_assets_dir, app_bundle, "index.js")
                    ):
                        bundles.append(app_bundle)

    webpack_stats = {
        "status": "done",
        "publicPath": "/static",
        "chunks": {},
        "assets": {},
    }
    for bundle in bundles:
        asset = f"js/{bundle}.js"
        webpack_stats["chunks"][bundle] = [asset]
        webpack_stats["assets"][asset] = {
            "name": asset,
            "publicPath": f"/static/{asset}",
        }

    webpack_load_assets = session_mocker.patch(
        "webpack_loader.loaders.WebpackLoader.load_assets"
    )
    webpack_load_assets.return_value = webpack_stats


@pytest.fixture(scope="function", autouse=True)
def random_seed(pytestconfig):
    state = random.getstate()
    seed = pytestconfig.getoption("--swh-web-random-seed")
    if seed is None:
        seed = time.time()
    seed = int(seed)
    cache.set(_random_seed_cache_key, seed)
    random.seed(seed)
    yield seed
    random.setstate(state)


def pytest_report_teststatus(report, *args):
    if report.when == "call" and report.outcome == "failed":
        seed = cache.get(_random_seed_cache_key, None)
        line = (
            f'FAILED {report.nodeid}: Use "pytest --swh-web-random-seed={seed} '
            f'{report.nodeid}" to reproduce that test failure with same inputs'
        )
        report.sections.append((_swh_web_custom_section, line))


def pytest_terminal_summary(terminalreporter, *args):
    reports = terminalreporter.getreports("failed")
    content = os.linesep.join(
        text
        for report in reports
        for secname, text in report.sections
        if secname == _swh_web_custom_section
    )
    if content:
        terminalreporter.ensure_newline()
        terminalreporter.section(_swh_web_custom_section, sep="-", blue=True, bold=True)
        terminalreporter.line(content)


@pytest.fixture
def django_cache():
    return cache


# Clear Django cache before each test
@pytest.fixture(autouse=True)
def django_cache_cleared(django_cache):
    django_cache.clear()


# Alias rf fixture from pytest-django
@pytest.fixture
def request_factory(rf):
    return rf


# Fixture to get test client from Django REST Framework
@pytest.fixture
def api_client():
    return APIClient()


# Fixture to get API request factory from Django REST Framework
@pytest.fixture
def api_request_factory():
    return APIRequestFactory()


# Initialize tests data
@pytest.fixture(scope="function", autouse=True)
def tests_data():
    data = get_tests_data(reset=True)
    # Update swh-web configuration to use the in-memory storages
    # instantiated in the tests.data module
    override_storages(
        data["storage"], data["idx_storage"], data["search"], data["counters"]
    )
    return data


@pytest.fixture(scope="function")
def sha1():
    """Fixture returning a valid hexadecimal sha1 value."""
    return random_sha1()


@pytest.fixture(scope="function")
def invalid_sha1():
    """Fixture returning an invalid sha1 representation."""
    return hash_to_hex(bytes(random.randint(0, 255) for _ in range(50)))


@pytest.fixture(scope="function")
def sha256():
    """Fixture returning a valid hexadecimal sha256 value."""
    return random_sha256()


def _known_swh_objects(tests_data, object_type):
    return tests_data[object_type]


@pytest.fixture(scope="function")
def content(tests_data):
    """Fixture returning a random content ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "contents"))


@pytest.fixture(scope="function")
def contents(tests_data):
    """Fixture returning random contents ingested into the test archive."""
    return random.choices(
        _known_swh_objects(tests_data, "contents"), k=random.randint(2, 8)
    )


def _new_content(tests_data):
    while True:
        new_content = random_content()
        sha1_bytes = hash_to_bytes(new_content["sha1"])
        if tests_data["storage"].content_get_data({"sha1": sha1_bytes}) is None:
            return new_content


@pytest.fixture(scope="function")
def unknown_content(tests_data):
    """Fixture returning a random content not ingested into the test archive."""
    return _new_content(tests_data)


@pytest.fixture(scope="function")
def unknown_contents(tests_data):
    """Fixture returning random contents not ingested into the test archive."""
    new_contents = []
    new_content_ids = set()
    nb_contents = random.randint(2, 8)
    while len(new_contents) != nb_contents:
        new_content = _new_content(tests_data)
        if new_content["sha1"] not in new_content_ids:
            new_contents.append(new_content)
            new_content_ids.add(new_content["sha1"])
    return list(new_contents)


@pytest.fixture(scope="function")
def empty_content():
    """Fixture returning the empty content ingested into the test archive."""
    empty_content = Content.from_data(data=b"").to_dict()
    for algo in DEFAULT_ALGORITHMS:
        empty_content[algo] = hash_to_hex(empty_content[algo])
    return empty_content


@functools.lru_cache(maxsize=None)
def _content_text():
    return list(
        filter(
            lambda c: c["mimetype"].startswith("text/"),
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_text():
    """
    Fixture returning a random textual content ingested into the test archive.
    """
    return random.choice(_content_text())


@functools.lru_cache(maxsize=None)
def _content_text_non_utf8():
    return list(
        filter(
            lambda c: c["mimetype"].startswith("text/")
            and c["encoding"] not in ("utf-8", "us-ascii"),
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_text_non_utf8():
    """Fixture returning a random textual content not encoded to UTF-8 ingested
    into the test archive.
    """
    return random.choice(_content_text_non_utf8())


@functools.lru_cache(maxsize=None)
def _content_application_no_highlight():
    return list(
        filter(
            lambda c: c["mimetype"].startswith("application/")
            and c["hljs_language"] == "plaintext",
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_application_no_highlight():
    """Fixture returning a random textual content with mimetype
    starting with application/ and no detected programming language to
    highlight ingested into the test archive.
    """
    return random.choice(_content_application_no_highlight())


@functools.lru_cache(maxsize=None)
def _content_text_no_highlight():
    return list(
        filter(
            lambda c: c["mimetype"].startswith("text/")
            and c["hljs_language"] == "plaintext",
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_text_no_highlight():
    """Fixture returning a random textual content with no detected
    programming language to highlight ingested into the test archive.
    """
    return random.choice(_content_text_no_highlight())


@functools.lru_cache(maxsize=None)
def _content_image_type():
    return list(
        filter(
            lambda c: c["mimetype"] in browsers_supported_image_mimes,
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_image_type():
    """Fixture returning a random image content ingested into the test archive."""
    return random.choice(_content_image_type())


@functools.lru_cache(maxsize=None)
def _content_binary_type():
    return list(
        filter(
            lambda c: c["mimetype"] == "application/octet-stream",
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_binary_type():
    """Fixture returning a random image content ingested into the test archive."""
    return random.choice(_content_binary_type())


@functools.lru_cache(maxsize=None)
def _content_unsupported_image_type_rendering():
    return list(
        filter(
            lambda c: c["mimetype"].startswith("image/")
            and c["mimetype"] not in browsers_supported_image_mimes,
            _known_swh_objects(get_tests_data(), "contents"),
        )
    )


@pytest.fixture(scope="function")
def content_unsupported_image_type_rendering():
    """Fixture returning a random image content ingested into the test archive that
    cannot be rendered by browsers.
    """
    return random.choice(_content_unsupported_image_type_rendering())


@functools.lru_cache(maxsize=None)
def _content_utf8_detected_as_binary():
    def utf8_binary_detected(content):
        if content["encoding"] != "binary":
            return False
        try:
            content["raw_data"].decode("utf-8")
        except Exception:
            return False
        else:
            return True

    return list(
        filter(utf8_binary_detected, _known_swh_objects(get_tests_data(), "contents"))
    )


@pytest.fixture(scope="function")
def content_utf8_detected_as_binary():
    """Fixture returning a random textual content detected as binary
    by libmagic while they are valid UTF-8 encoded files.
    """

    return random.choice(_content_utf8_detected_as_binary())


@pytest.fixture(scope="function")
def directory(tests_data):
    """Fixture returning a random directory ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "directories"))


def _directory_contains_entry_type(tests_data, dir_id, entry_type):
    return any(
        [
            e["type"] == entry_type
            for e in list(tests_data["storage"].directory_ls(hash_to_bytes(dir_id)))
        ]
    )


@functools.lru_cache(maxsize=None)
def _directory_with_entry_type(type_):
    tests_data = get_tests_data()
    return list(
        filter(
            lambda d: _directory_contains_entry_type(tests_data, d, type_),
            _known_swh_objects(tests_data, "directories"),
        )
    )


@pytest.fixture(scope="function")
def directory_with_subdirs():
    """Fixture returning a random directory containing sub directories ingested
    into the test archive.
    """
    return random.choice(_directory_with_entry_type("dir"))


@pytest.fixture(scope="function")
def directory_with_files():
    """Fixture returning a random directory containing at least one regular file."""
    return random.choice(_directory_with_entry_type("file"))


@pytest.fixture(scope="function")
def revision_with_files_in_target_directory(tests_data):
    """Fixture returning a random directory containing at least one regular file."""
    storage = tests_data["storage"]
    return random.choice(
        list(
            filter(
                lambda r: _directory_contains_entry_type(
                    tests_data,
                    storage.revision_get([hash_to_bytes(r)])[0].directory.hex(),
                    "file",
                ),
                _known_swh_objects(tests_data, "revisions"),
            )
        )
    )


@pytest.fixture(scope="function")
def unknown_directory(tests_data):
    """Fixture returning a random directory not ingested into the test archive."""
    while True:
        new_directory = random_sha1()
        sha1_bytes = hash_to_bytes(new_directory)
        if list(tests_data["storage"].directory_missing([sha1_bytes])):
            return new_directory


@pytest.fixture(scope="function")
def empty_directory():
    """Fixture returning the empty directory ingested into the test archive."""
    return Directory(entries=()).id.hex()


@pytest.fixture(scope="function")
def revision(tests_data):
    """Fixturereturning a random revision ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "revisions"))


@pytest.fixture(scope="function")
def revisions(tests_data):
    """Fixture returning random revisions ingested into the test archive."""
    return random.choices(
        _known_swh_objects(tests_data, "revisions"),
        k=random.randint(2, 8),
    )


@pytest.fixture(scope="function")
def revisions_list(tests_data):
    """Fixture returning random revisions ingested into the test archive."""

    def gen_revisions_list(size):
        return random.choices(
            _known_swh_objects(tests_data, "revisions"),
            k=size,
        )

    return gen_revisions_list


@pytest.fixture(scope="function")
def unknown_revision(tests_data):
    """Fixture returning a random revision not ingested into the test archive."""
    while True:
        new_revision = random_sha1()
        sha1_bytes = hash_to_bytes(new_revision)
        if tests_data["storage"].revision_get([sha1_bytes])[0] is None:
            return new_revision


def _get_origin_dfs_revisions_walker(tests_data):
    storage = tests_data["storage"]
    origin = random.choice(tests_data["origins"][0:1])
    snapshot = snapshot_get_latest(storage, origin["url"])
    if snapshot.branches[b"HEAD"].target_type.value == "alias":
        target = snapshot.branches[b"HEAD"].target
        head = snapshot.branches[target].target
    else:
        head = snapshot.branches[b"HEAD"].target
    return get_revisions_walker("dfs", storage, head)


@functools.lru_cache(maxsize=None)
def _ancestor_revisions_data():
    # get a dfs revisions walker for one of the origins
    # loaded into the test archive
    revisions_walker = _get_origin_dfs_revisions_walker(get_tests_data())
    master_revisions = []
    children = defaultdict(list)
    init_rev_found = False
    # TODO: remove this when the migration to new revision_walker/revision_log
    # is done in swh-storage (so probably when swh-storage v3 is published)
    revisions = (
        Revision.from_dict(rev) if isinstance(rev, dict) else rev
        for rev in revisions_walker
    )
    # get revisions only authored in the master branch
    for rev in revisions:
        for rev_p in rev.parents:
            children[rev_p].append(rev.id)
        if not init_rev_found:
            master_revisions.append(rev)
        if not rev.parents:
            init_rev_found = True
    return master_revisions, children


@pytest.fixture(scope="function")
def ancestor_revisions():
    """Fixture returning a pair of revisions ingested into the test archive
    with an ancestor relation.
    """
    master_revisions, children = _ancestor_revisions_data()
    # head revision
    root_rev = master_revisions[0]
    # pick a random revision, different from head, only authored
    # in the master branch
    ancestor_rev_idx = random.choice(list(range(1, len(master_revisions) - 1)))
    ancestor_rev = master_revisions[ancestor_rev_idx]
    ancestor_child_revs = children[ancestor_rev.id]

    return {
        "sha1_git_root": hash_to_hex(root_rev.id),
        "sha1_git": hash_to_hex(ancestor_rev.id),
        "children": [hash_to_hex(r) for r in ancestor_child_revs],
    }


@functools.lru_cache(maxsize=None)
def _non_ancestor_revisions_data():
    # get a dfs revisions walker for one of the origins
    # loaded into the test archive
    revisions_walker = _get_origin_dfs_revisions_walker(get_tests_data())
    merge_revs = []
    children = defaultdict(list)
    # TODO: remove this when the migration to new revision_walker/revision_log
    # is done in swh-storage (so probably when swh-storage v3 is published)
    revisions = (
        Revision.from_dict(rev) if isinstance(rev, dict) else rev
        for rev in revisions_walker
    )
    # get all merge revisions
    for rev in revisions:
        if len(rev.parents) > 1:
            merge_revs.append(rev)
        for rev_p in rev.parents:
            children[rev_p].append(rev.id)
    return merge_revs, children


@pytest.fixture(scope="function")
def non_ancestor_revisions():
    """Fixture returning a pair of revisions ingested into the test archive
    with no ancestor relation.
    """
    merge_revs, children = _non_ancestor_revisions_data()
    # find a merge revisions whose parents have a unique child revision
    random.shuffle(merge_revs)
    selected_revs = None
    for merge_rev in merge_revs:
        if all(len(children[rev_p]) == 1 for rev_p in merge_rev.parents):
            selected_revs = merge_rev.parents

    return {
        "sha1_git_root": hash_to_hex(selected_revs[0]),
        "sha1_git": hash_to_hex(selected_revs[1]),
    }


@pytest.fixture(scope="function")
def revision_with_submodules():
    """Fixture returning a revision that is known to
    point to a directory with revision entries (aka git submodules)
    """

    return {
        "rev_sha1_git": "ffcb69001f3f6745dfd5b48f72ab6addb560e234",
        "rev_dir_sha1_git": "d92a21446387fa28410e5a74379c934298f39ae2",
        "rev_dir_rev_path": "libtess2",
    }


def _get_origin(origin_url):
    origins = _known_swh_objects(get_tests_data(), "origins")
    return next(
        filtered_origin
        for filtered_origin in origins
        if filtered_origin["url"] == origin_url
    )


@pytest.fixture(scope="function")
def origin_with_metadata_file():
    """Fixture returning the origin with raw intrinsic metadata (codemeta.json
    and citation.cff file) ingested into the test archive."""
    return _get_origin(ORIGIN_WITH_METADATA_FILES)


@pytest.fixture(scope="function")
def origin_with_cff_file():
    """Fixture returning the origin with raw intrinsic metadata (codemeta.json
    and citation.cff file) ingested into the test archive."""
    return _get_origin(ORIGIN_WITH_CFF_FILE)


@pytest.fixture(scope="function")
def origin_url_with_unknown_codemeta_context():
    """Fixture returning the origin with raw intrinsic metadata (codemeta.json
    and citation.cff file) ingested into the test archive."""
    return ORIGIN_WITH_UNKNOWN_CODEMETA_CONTEXT


@pytest.fixture(scope="function")
def objects_with_metadata_file(tests_data):
    """Fixture returning the objects (snapshot and revision)
    with raw intrinsic metadata (codemeta.json and citation.cff file)
    ingested into the test archive."""
    swh_objects = _known_swh_objects(get_tests_data(), "swhids")
    snapshot = snapshot_get_latest(tests_data["storage"], ORIGIN_WITH_METADATA_FILES)
    snapshot_branch = snapshot_resolve_alias(
        tests_data["storage"], hash_to_bytes(snapshot.id), b"HEAD"
    )
    object_ids_with_metadata_file = [snapshot.id, snapshot_branch.target]
    return [
        object_with_metadata_file
        for object_with_metadata_file in swh_objects
        if (
            object_with_metadata_file.object_id in object_ids_with_metadata_file
            or (
                object_with_metadata_file.anchor is not None
                and object_with_metadata_file.anchor.object_id
                in object_ids_with_metadata_file
            )
        )
    ]


@pytest.fixture(scope="function")
def release(tests_data):
    """Fixture returning a random release ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "releases"))


@pytest.fixture(scope="function")
def releases(tests_data):
    """Fixture returning random releases ingested into the test archive."""
    return random.choices(
        _known_swh_objects(tests_data, "releases"), k=random.randint(2, 8)
    )


@pytest.fixture(scope="function")
def unknown_release(tests_data):
    """Fixture returning a random release not ingested into the test archive."""
    while True:
        new_release = random_sha1()
        sha1_bytes = hash_to_bytes(new_release)
        if tests_data["storage"].release_get([sha1_bytes])[0] is None:
            return new_release


@pytest.fixture(scope="function")
def snapshot(tests_data):
    """Fixture returning a random snapshot ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "snapshots"))


@pytest.fixture(scope="function")
def unknown_snapshot(tests_data):
    """Fixture returning a random snapshot not ingested into the test archive."""
    while True:
        new_snapshot = random_sha1()
        sha1_bytes = hash_to_bytes(new_snapshot)
        if tests_data["storage"].snapshot_get_branches(sha1_bytes) is None:
            return new_snapshot


@pytest.fixture(scope="function")
def origin(tests_data):
    """Fixture returning a random origin ingested into the test archive."""
    return random.choice(_known_swh_objects(tests_data, "origins"))


@functools.lru_cache(maxsize=None)
def _origin_with_multiple_visits():
    tests_data = get_tests_data()
    origins = []
    storage = tests_data["storage"]
    for origin in tests_data["origins"]:
        visit_page = storage.origin_visit_get(origin["url"])
        if len(visit_page.results) > 1:
            origins.append(origin)
    return origins


@pytest.fixture(scope="function")
def origin_with_multiple_visits():
    """Fixture returning a random origin with multiple visits ingested
    into the test archive.
    """
    return random.choice(_origin_with_multiple_visits())


@functools.lru_cache(maxsize=None)
def _origin_with_releases():
    tests_data = get_tests_data()
    origins = []
    for origin in tests_data["origins"]:
        snapshot = snapshot_get_latest(tests_data["storage"], origin["url"])
        if any([b.target_type.value == "release" for b in snapshot.branches.values()]):
            origins.append(origin)
    return origins


@pytest.fixture(scope="function")
def origin_with_releases():
    """Fixture returning a random origin with releases ingested into the test archive."""
    return random.choice(_origin_with_releases())


@functools.lru_cache(maxsize=None)
def _origin_with_pull_request_branches():
    tests_data = get_tests_data()
    origins = []
    storage = tests_data["storage"]
    for origin in storage.origin_list(limit=1000).results:
        snapshot = snapshot_get_latest(storage, origin.url)
        if snapshot and any([b"refs/pull/" in b for b in snapshot.branches]):
            origins.append(origin)
    return origins


@pytest.fixture(scope="function")
def origin_with_pull_request_branches():
    """Fixture returning a random origin with pull request branches ingested
    into the test archive.
    """
    return random.choice(_origin_with_pull_request_branches())


@functools.lru_cache(maxsize=None)
def _origins_with_non_empty_snapshot():
    tests_data = get_tests_data()
    origins = []
    storage = tests_data["storage"]
    for origin in storage.origin_list(limit=1000).results:
        snapshot = snapshot_get_latest(storage, origin.url)
        if snapshot and snapshot.branches:
            origins.append(origin)
    return origins


@pytest.fixture
def origins_with_non_empty_snapshot():
    return _origins_with_non_empty_snapshot()


@pytest.fixture
def origin_with_multiple_visit_types(archive_data, tests_data):
    return "https://example.org/project/multiple/visit/types"


@pytest.fixture
def origin_with_quoted_space_in_url():
    return ORIGIN_WITH_QUOTED_SPACE_IN_URL


@functools.lru_cache(maxsize=None)
def _object_type_swhid(object_type):
    return list(
        filter(
            lambda swhid: swhid.object_type == object_type,
            _known_swh_objects(get_tests_data(), "swhids"),
        )
    )


@pytest.fixture(scope="function")
def content_swhid():
    """Fixture returning a qualified SWHID for a random content object
    ingested into the test archive.
    """
    return random.choice(_object_type_swhid(ObjectType.CONTENT))


@pytest.fixture(scope="function")
def directory_swhid():
    """Fixture returning a qualified SWHID for a random directory object
    ingested into the test archive.
    """
    return random.choice(_object_type_swhid(ObjectType.DIRECTORY))


@pytest.fixture(scope="function")
def release_swhid():
    """Fixture returning a qualified SWHID for a random release object
    ingested into the test archive.
    """
    return random.choice(_object_type_swhid(ObjectType.RELEASE))


@pytest.fixture(scope="function")
def revision_swhid():
    """Fixture returning a qualified SWHID for a random revision object
    ingested into the test archive.
    """
    return random.choice(_object_type_swhid(ObjectType.REVISION))


@pytest.fixture(scope="function")
def snapshot_swhid():
    """Fixture returning a qualified SWHID for a snapshot object
    ingested into the test archive.
    """
    return random.choice(_object_type_swhid(ObjectType.SNAPSHOT))


@pytest.fixture(scope="function", params=list(ObjectType))
def unknown_core_swhid(request) -> CoreSWHID:
    """Fixture returning an unknown core SWHID.

    Tests using this will be called once per object type.
    """
    return CoreSWHID(
        object_type=request.param,
        object_id=random_sha1_bytes(),
    )


# Fixture to manipulate data from a sample archive used in the tests
@pytest.fixture(scope="function")
def archive_data(tests_data):
    return _ArchiveData(tests_data)


# Fixture to manipulate indexer data from a sample archive used in the tests
@pytest.fixture(scope="function")
def indexer_data(tests_data):
    return _IndexerData(tests_data)


class _ArchiveData:
    """
    Helper class to manage data from a sample test archive.

    It is initialized with a reference to an in-memory storage
    containing raw tests data.

    It is basically a proxy to Storage interface but it overrides some methods
    to retrieve those tests data in a json serializable format in order to ease
    tests implementation.
    """

    def __init__(self, tests_data):
        self.storage = tests_data["storage"]

    def __getattr__(self, key):
        if key == "storage":
            raise AttributeError(key)
        # Forward calls to non overridden Storage methods to wrapped
        # storage instance
        return getattr(self.storage, key)

    def content_find(self, content: Dict[str, Any]) -> Dict[str, Any]:
        cnt_ids_bytes = {
            algo_hash: hash_to_bytes(content[algo_hash])
            for algo_hash in ALGORITHMS
            if content.get(algo_hash)
        }
        cnt = self.storage.content_find(cnt_ids_bytes)
        return converters.from_content(cnt[0].to_dict()) if cnt else cnt

    def content_get(self, cnt_id: str) -> Dict[str, Any]:
        cnt_id_bytes = hash_to_bytes(cnt_id)
        content = self.storage.content_get([cnt_id_bytes])[0]
        if content:
            content_d = content.to_dict()
            content_d.pop("ctime", None)
        else:
            content_d = None
        return converters.from_swh(
            content_d, hashess={"sha1", "sha1_git", "sha256", "blake2s256"}
        )

    def content_get_data(self, cnt_id: str) -> Optional[Dict[str, Any]]:
        cnt_id_bytes = hash_to_bytes(cnt_id)
        cnt_data = self.storage.content_get_data({"sha1": cnt_id_bytes})
        if cnt_data is None:
            return None
        return converters.from_content({"data": cnt_data, "sha1": cnt_id_bytes})

    def directory_get(self, dir_id):
        return {"id": dir_id, "content": self.directory_ls(dir_id)}

    def directory_ls(self, dir_id, recursive=False):
        cnt_id_bytes = hash_to_bytes(dir_id)
        dir_content = map(
            converters.from_directory_entry,
            self.storage.directory_ls(cnt_id_bytes, recursive=recursive),
        )
        return list(dir_content)

    def release_get(self, rel_id: str) -> Optional[Dict[str, Any]]:
        rel_id_bytes = hash_to_bytes(rel_id)
        rel_data = self.storage.release_get([rel_id_bytes])[0]
        return converters.from_release(rel_data) if rel_data else None

    def revision_get(self, rev_id: str) -> Optional[Dict[str, Any]]:
        rev_id_bytes = hash_to_bytes(rev_id)
        rev_data = self.storage.revision_get([rev_id_bytes])[0]
        return converters.from_revision(rev_data) if rev_data else None

    def revision_log(self, rev_id, limit=None):
        rev_id_bytes = hash_to_bytes(rev_id)
        return list(
            map(
                converters.from_revision,
                self.storage.revision_log([rev_id_bytes], limit=limit),
            )
        )

    def snapshot_get_latest(self, origin_url):
        snp = snapshot_get_latest(self.storage, origin_url)
        return converters.from_snapshot(snp.to_dict())

    def origin_get(self, origin_urls):
        origins = self.storage.origin_get(origin_urls)
        return [converters.from_origin(o.to_dict()) for o in origins]

    def origin_visit_get(self, origin_url):
        next_page_token = None
        visits = []
        while True:
            visit_page = self.storage.origin_visit_get(
                origin_url, page_token=next_page_token
            )
            next_page_token = visit_page.next_page_token

            for visit in visit_page.results:
                visit_status = self.storage.origin_visit_status_get_latest(
                    origin_url, visit.visit
                )
                visits.append(converters.from_origin_visit(visit, visit_status))
            if not next_page_token:
                break
        return visits

    def origin_visit_get_by(
        self, origin_url: str, visit_id: int
    ) -> Optional[OriginVisitInfo]:
        visit = self.storage.origin_visit_get_by(origin_url, visit_id)
        visit_status = self.storage.origin_visit_status_get_latest(origin_url, visit_id)
        return converters.from_origin_visit(visit, visit_status)

    def origin_visit_status_get_latest(
        self,
        origin_url,
        type: Optional[str] = None,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
    ):
        visit = self.storage.origin_visit_get_latest(
            origin_url,
            type=type,
            allowed_statuses=allowed_statuses,
            require_snapshot=require_snapshot,
        )
        visit_status = self.storage.origin_visit_status_get_latest(
            origin_url,
            visit.visit,
            allowed_statuses=allowed_statuses,
            require_snapshot=require_snapshot,
        )
        return (
            converters.from_origin_visit(visit, visit_status) if visit_status else None
        )

    def snapshot_get(self, snapshot_id):
        snp = snapshot_get_all_branches(self.storage, hash_to_bytes(snapshot_id))
        return converters.from_snapshot(snp.to_dict())

    def snapshot_get_branches(
        self, snapshot_id, branches_from="", branches_count=1000, target_types=None
    ):
        partial_branches = self.storage.snapshot_get_branches(
            hash_to_bytes(snapshot_id),
            branches_from.encode(),
            branches_count,
            target_types,
        )
        return converters.from_partial_branches(partial_branches)

    def snapshot_get_head(self, snapshot):
        if snapshot["branches"]["HEAD"]["target_type"] == "alias":
            target = snapshot["branches"]["HEAD"]["target"]
            head = snapshot["branches"][target]["target"]
        else:
            head = snapshot["branches"]["HEAD"]["target"]
        return head

    def snapshot_count_branches(self, snapshot_id):
        counts = dict.fromkeys(("alias", "release", "revision"), 0)
        counts.update(self.storage.snapshot_count_branches(hash_to_bytes(snapshot_id)))
        counts.pop(None, None)
        counts["branch"] = sum(
            counts.get(target_type, 0)
            for target_type in ("content", "directory", "revision")
        )
        return counts


class _IndexerData:
    """
    Helper class to manage indexer tests data

    It is initialized with a reference to an in-memory indexer storage
    containing raw tests data.

    It also defines class methods to retrieve those tests data in
    a json serializable format in order to ease tests implementation.

    """

    def __init__(self, tests_data):
        self.idx_storage = tests_data["idx_storage"]
        self.mimetype_indexer = tests_data["mimetype_indexer"]
        self.license_indexer = tests_data["license_indexer"]

    def content_add_mimetype(self, cnt_id):
        self.mimetype_indexer.run([objid_from_dict({"sha1": hash_to_bytes(cnt_id)})])

    def content_get_mimetype(self, cnt_id):
        mimetype = self.idx_storage.content_mimetype_get([hash_to_bytes(cnt_id)])[
            0
        ].to_dict()
        return converters.from_filetype(mimetype)

    def content_add_license(self, cnt_id):
        self.license_indexer.run([objid_from_dict({"sha1": hash_to_bytes(cnt_id)})])

    def content_get_license(self, cnt_id):
        cnt_id_bytes = hash_to_bytes(cnt_id)
        licenses = self.idx_storage.content_fossology_license_get([cnt_id_bytes])
        for license in licenses:
            yield converters.from_swh(license.to_dict(), hashess={"id"})


@pytest.fixture
def keycloak_oidc(keycloak_oidc, mocker):
    keycloak_config = get_config()["keycloak"]

    keycloak_oidc.server_url = keycloak_config["server_url"]
    keycloak_oidc.realm_name = keycloak_config["realm_name"]
    keycloak_oidc.client_id = settings.OIDC_SWH_WEB_CLIENT_ID

    keycloak_oidc_client = mocker.patch("swh.web.auth.views.keycloak_oidc_client")
    keycloak_oidc_client.return_value = keycloak_oidc

    return keycloak_oidc


@pytest.fixture
def subtest(request):
    """A hack to explicitly set up and tear down fixtures.

    This fixture allows you to set up and tear down fixtures within the test
    function itself. This is useful (necessary!) for using Hypothesis inside
    pytest, as hypothesis will call the test function multiple times, without
    setting up or tearing down fixture state as it is normally the case.

    Copied from the pytest-subtesthack project, public domain license
    (https://github.com/untitaker/pytest-subtesthack).
    """
    parent_test = request.node

    def inner(func):
        if hasattr(Function, "from_parent"):
            item = Function.from_parent(
                parent_test,
                name=request.function.__name__ + "[]",
                originalname=request.function.__name__,
                callobj=func,
            )
        else:
            item = Function(
                name=request.function.__name__ + "[]", parent=parent_test, callobj=func
            )
        nextitem = parent_test  # prevents pytest from tearing down module fixtures

        item.ihook.pytest_runtest_setup(item=item)
        try:
            item.ihook.pytest_runtest_call(item=item)
        finally:
            item.ihook.pytest_runtest_teardown(item=item, nextitem=nextitem)

    return inner


@pytest.fixture
def swh_scheduler(config_updater):
    from swh.scheduler import get_scheduler
    from swh.scheduler.model import TaskType

    swh_scheduler = get_scheduler(cls="memory")
    config_updater({"scheduler": swh_scheduler})
    # create load-git and load-hg task types
    for task_type in TASK_TYPES.values():
        # see https://gitlab.softwareheritage.org/swh/devel/swh-scheduler/-/commit/c46ffadf7adf24c7eb3ffce062e8ade3818c79cc  # noqa
        swh_scheduler.create_task_type(
            task_type.evolve(type=task_type.type.replace("load-test-", "load-", 1))
        )
    # create load-svn task type
    swh_scheduler.create_task_type(
        TaskType(
            type="load-svn",
            description="Update a Subversion repository",
            backend_name="swh.loader.svn.tasks.DumpMountAndLoadSvnRepository",
            default_interval=timedelta(days=64),
            min_interval=timedelta(hours=12),
            max_interval=timedelta(days=64),
            backoff_factor=2.0,
            max_queue_length=None,
            num_retries=7,
            retry_delay=timedelta(hours=2),
        )
    )
    # create load-cvs task type
    swh_scheduler.create_task_type(
        TaskType(
            type="load-cvs",
            description="Update a CVS repository",
            backend_name="swh.loader.cvs.tasks.DumpMountAndLoadSvnRepository",
            default_interval=timedelta(days=64),
            min_interval=timedelta(hours=12),
            max_interval=timedelta(days=64),
            backoff_factor=2.0,
            max_queue_length=None,
            num_retries=7,
            retry_delay=timedelta(hours=2),
        )
    )
    # create load-bzr task type
    swh_scheduler.create_task_type(
        TaskType(
            type="load-bzr",
            description="Update a Bazaar repository",
            backend_name="swh.loader.bzr.tasks.LoadBazaar",
            default_interval=timedelta(days=64),
            min_interval=timedelta(hours=12),
            max_interval=timedelta(days=64),
            backoff_factor=2.0,
            max_queue_length=None,
            num_retries=7,
            retry_delay=timedelta(hours=2),
        )
    )

    # create load-bzr task type
    swh_scheduler.create_task_type(
        TaskType(
            type="load-tarball-directory",
            description="Load a tarball into the archive",
            backend_name="swh.loader.core.tasks.LoadTarballDirectory",
            default_interval=timedelta(days=64),
            min_interval=timedelta(hours=12),
            max_interval=timedelta(days=64),
            backoff_factor=2.0,
            max_queue_length=None,
            num_retries=7,
            retry_delay=timedelta(hours=2),
        )
    )

    # add method to add load-archive-files task type during tests
    def add_load_archive_task_type():
        swh_scheduler.create_task_type(
            TaskType(
                type="load-archive-files",
                description="Load tarballs",
                backend_name="swh.loader.package.archive.tasks.LoadArchive",
                default_interval=timedelta(days=64),
                min_interval=timedelta(hours=12),
                max_interval=timedelta(days=64),
                backoff_factor=2.0,
                max_queue_length=None,
                num_retries=7,
                retry_delay=timedelta(hours=2),
            )
        )

    # create list-save-bulk task type
    swh_scheduler.create_task_type(
        TaskType(
            type="list-save-bulk",
            description="Check a list of origins provided by a user",
            backend_name="swh.lister.save_bulk.tasks.SaveBulkListerTask",
            default_interval=timedelta(days=64),
            min_interval=timedelta(hours=12),
            max_interval=timedelta(days=64),
            backoff_factor=2.0,
            max_queue_length=None,
            num_retries=7,
            retry_delay=timedelta(hours=2),
        )
    )

    swh_scheduler.add_load_archive_task_type = add_load_archive_task_type

    yield swh_scheduler
    get_scheduler_load_task_types.cache_clear()


@pytest.fixture(scope="session")
def django_db_setup(request, django_db_blocker, postgresql_proc):
    from django.conf import settings

    settings.DATABASES["default"].update(
        {
            ("ENGINE", "django.db.backends.postgresql"),
            ("NAME", get_config()["test_db"]["name"]),
            ("USER", postgresql_proc.user),
            ("HOST", postgresql_proc.host),
            ("PORT", postgresql_proc.port),
        }
    )
    with django_db_blocker.unblock():
        setup_databases(
            verbosity=request.config.option.verbose, interactive=False, keepdb=False
        )


@pytest.fixture
def staff_user():
    return User.objects.create_user(username="admin", password="", is_staff=True)


@pytest.fixture
def regular_user():
    return User.objects.create_user(username="johndoe", password="")


@pytest.fixture
def regular_user2():
    return User.objects.create_user(username="janedoe", password="")


@pytest.fixture
def add_forge_moderator():
    moderator = User.objects.create_user(username="add-forge moderator", password="")
    moderator.user_permissions.add(
        get_or_create_django_permission(ADD_FORGE_MODERATOR_PERMISSION)
    )
    return moderator


@pytest.fixture
def mailmap_admin():
    mailmap_admin = User.objects.create_user(username="mailmap-admin", password="")
    mailmap_admin.user_permissions.add(
        get_or_create_django_permission(MAILMAP_ADMIN_PERMISSION)
    )
    return mailmap_admin


@pytest.fixture
def mailmap_user():
    mailmap_user = User.objects.create_user(username="mailmap-user", password="")
    mailmap_user.user_permissions.add(
        get_or_create_django_permission(MAILMAP_PERMISSION)
    )
    return mailmap_user


@pytest.fixture
def save_bulk_user():
    save_bulk_user = User.objects.create_user(username="save-bulk-user", password="")
    save_bulk_user.user_permissions.add(
        get_or_create_django_permission(API_SAVE_BULK_PERMISSION)
    )
    return save_bulk_user


@pytest.fixture
def provenance_user():
    provenance_user = User.objects.create_user(username="provenance-user", password="")
    provenance_user.user_permissions.add(
        get_or_create_django_permission(API_PROVENANCE_PERMISSION)
    )
    return provenance_user


@pytest.fixture
def alter_admin():
    alter_admin = User.objects.create_user(username="alter-admin", password="")
    alter_admin.user_permissions.add(
        get_or_create_django_permission(ALTER_ADMIN_PERMISSION)
    )
    return alter_admin


def reload_urlconf():
    from django.conf import settings

    clear_url_caches()
    # force reloading of all URLs as they depend on django settings
    # and swh-web configuration
    urlconfs = [settings.ROOT_URLCONF]
    urlconfs += [f"{app}.urls" for app in settings.SWH_DJANGO_APPS]
    for urlconf in urlconfs:
        try:
            if urlconf in sys.modules:
                reload(sys.modules[urlconf])
            else:
                import_module(urlconf)
        except ModuleNotFoundError:
            pass


class SwhSettingsWrapper(SettingsWrapper):
    def __setattr__(self, attr: str, value) -> None:
        super().__setattr__(attr, value)
        reload_urlconf()

    def finalize(self) -> None:
        super().finalize()
        reload_urlconf()


@pytest.fixture
def django_settings():
    """Override pytest-django settings fixture in order to reload URLs
    when modifying settings in test and after test execution as most of
    them depend on installed django apps in swh-web.
    """
    settings = SwhSettingsWrapper()
    yield settings
    settings.finalize()


def _reload_django_settings():
    reload(common_settings)
    reload(tests_settings)
    settings._setup()
    finders.get_finder.cache_clear()
    get_app_template_dirs.cache_clear()
    reload_urlconf()
    # ensure to reload templates config
    engines._templates = None
    engines._engines = {}
    if hasattr(engines, "templates"):
        del engines.templates


@pytest.fixture
def config_updater(mocker):
    """Temporarily modify swh-web configuration and reload django settings."""
    original_config = swhweb_config.get_config()
    modules_to_patch = [
        module
        for _, module in filter(
            lambda m: m[0].startswith("swh.web.") and hasattr(m[1], "get_config"),
            sys.modules.items(),
        )
    ]

    def update_config(new_config):
        patched_config = dict(original_config)
        patched_config.update(new_config)
        mocker.patch.object(swhweb_config, "swhweb_config", dict(patched_config))
        _reload_django_settings()
        for module in modules_to_patch:
            mocker.patch.object(module, "get_config").return_value = dict(
                patched_config
            )

    yield update_config

    mocker.patch.object(swhweb_config, "swhweb_config", original_config)
    for module in modules_to_patch:
        mocker.patch.object(module, "get_config").return_value = original_config
    _reload_django_settings()


@pytest.fixture
def patch_backend(mocker):
    """Fixture to patch a swh.web.config backend instance

    Example:

      def test_xxx(patch_backend):
        mock = patch_backend("search", "origin_search", return_value=XXX)
        # ...
        assert mock.assert_called_once()
    """
    from swh.web import config

    def mock_backend(backend, attr, *args, **kw):
        obj = config.get_config()[backend]
        return mocker.patch.object(obj, attr, *args, **kw)

    return mock_backend



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Tuple

from rest_framework.request import Request

from swh.web.save_origin_webhooks.generic_receiver import OriginSaveWebhookReceiver


class GitHubOriginSaveWebhookReceiver(OriginSaveWebhookReceiver):
    FORGE_TYPE = "GitHub"
    WEBHOOK_GUIDE_URL = (
        "https://docs.github.com/en/developers/webhooks-and-events/"
        "webhooks/creating-webhooks#setting-up-a-webhook"
    )
    REPO_TYPES = "git"

    def is_forge_request(self, request: Request) -> bool:
        return (
            request.headers.get("User-Agent", "").startswith(
                f"{self.FORGE_TYPE}-Hookshot/"
            )
            and f"X-{self.FORGE_TYPE}-Event" in request.headers
        )

    def is_ping_event(self, request: Request) -> bool:
        return request.headers[f"X-{self.FORGE_TYPE}-Event"] == "ping"

    def is_push_event(self, request: Request) -> bool:
        return request.headers[f"X-{self.FORGE_TYPE}-Event"] == "push"

    def extract_repo_info(self, request: Request) -> Tuple[str, str, bool]:
        repo_url = request.data.get("repository", {}).get("html_url", "")
        private = request.data.get("repository", {}).get("private", False)
        return repo_url, "git", private


api_origin_save_webhook_github = GitHubOriginSaveWebhookReceiver()



# Copyright (C) 2021-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from django.core.management.base import BaseCommand

from swh.web.save_code_now.origin_save import (
    refresh_save_origin_request_statuses,
    schedule_origins_recurrent_visits,
)


class Command(BaseCommand):
    help = "Refresh save code now origin request statuses periodically"

    def handle(self, *args, **options):
        """Refresh origin save code now requests.

        For the origin visit types, svn, git, hg, this also installs the origins as
        recurring origins to visit.

        """
        updated_requests = refresh_save_origin_request_statuses()

        # then schedule the origins with meaningful status and type to be ingested
        # regularly
        nb_origins_scheduled = schedule_origins_recurrent_visits(updated_requests)

        if len(updated_requests) > 0:
            msg = f"Successfully updated {len(updated_requests)} save request(s)."
            if nb_origins_scheduled:
                msg += (
                    f"\n{nb_origins_scheduled} origins were also scheduled "
                    "for recurrent visits."
                )
        else:
            msg = "Nothing to do."

        self.stdout.write(self.style.SUCCESS(msg))



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Fixtures for e2e tests related to swh-web-alter."""

from datetime import timedelta

from django.utils import timezone

from swh.web.alter.models import (
    Alteration,
    AlterationCategory,
    AlterationStatus,
    BlockList,
    Event,
    EventCategory,
    EventRecipient,
    Origin,
    Token,
)

alterations = {
    # A copyright alteration request
    "00000000-0000-4000-8000-000000000001": {
        "status": AlterationStatus.PLANNING,
        "category": AlterationCategory.COPYRIGHT,
        "reasons": "not published under an open license",
        "expected_outcome": "delete everything",
        "email": "user1@domain.local",
        "origins": [
            "https://gitlab.local/user1/code",
            "https://gitlab.local/user1/project",
        ],
    }
}
for id_, a_props in alterations.items():
    origins = a_props.pop("origins")
    alteration, _ = Alteration.objects.get_or_create(id=id_, **a_props)  # type: ignore
    for url in origins:
        Origin.objects.get_or_create(url=url, alteration=alteration)

events = {
    # Events for the copyright alteration request
    "00000000-0000-4000-9000-000000000001": {
        "category": EventCategory.LOG,
        "alteration_id": "00000000-0000-4000-8000-000000000001",
        "content": "created",
        "internal": False,
    },
    "00000000-0000-4000-9000-000000000002": {
        "category": EventCategory.MESSAGE,
        "alteration_id": "00000000-0000-4000-8000-000000000001",
        "author": "Requester",
        "recipient": EventRecipient.SUPPORT,
        "content": "I would like to be informed of the progress of my request",
        "internal": False,
    },
    "00000000-0000-4000-9000-000000000003": {
        "category": EventCategory.MESSAGE,
        "alteration_id": "00000000-0000-4000-8000-000000000001",
        "author": EventRecipient.SUPPORT,
        "recipient": EventRecipient.LEGAL,
        "content": "Please check this internal message",
        "internal": True,
    },
}

for id_, e_props in events.items():
    Event.objects.get_or_create(id=id_, **e_props)  # type: ignore

# an expired access token for the copyright alteration request
Token.objects.get_or_create(
    value="ExpiredAccessToken",
    alteration_id="00000000-0000-4000-8000-000000000001",
    defaults={"expires_at": timezone.now() - timedelta(days=1)},
)
# an expired email confirmation token
Token.objects.get_or_create(
    value="ExpiredEmailToken",
    email="expired@domain.local",
    defaults={"expires_at": timezone.now() - timedelta(days=1)},
)
# an active access token for the copyright alteration request
Token.objects.get_or_create(
    value="ValidAccessToken",
    alteration_id="00000000-0000-4000-8000-000000000001",
    defaults={"expires_at": timezone.now() + timedelta(days=365)},
)

# block blocked@domain.local email
BlockList.objects.get_or_create(email_or_domain="blocked@domain.local", reason="spam")
